!> \file
!! Contains subroutines conduction_imp_gc(),
!!  Mat_prod_con() and cmp_precond_con()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine CONDUCTION_IMP_GC
!
!> Computes the implicit thermal gas conduction using the
!! conjugate gradient method.
!<
subroutine conduction_imp_gc
  
  use var
  use geom
  use divers    , only : verbose,mu,dt
  use parameters
  use cl_con
  use para      , only : mype,ncpu
  use unites    , only : an
  use conduction

  implicit none

#if PARA==1
  include 'mpif.h'
  integer :: info
#endif

  integer :: i,j,k,nite
  
  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: TT,TTold
  real :: dif,varmax,var_loc,xx,temp

  interface
     subroutine mat_prod_con(X,AX)
       use parameters, only : nxmin, nxmax
       implicit none
       real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: X, AX
     end subroutine mat_prod_con
     subroutine cmp_precond_con(R,Z)
       use parameters, only : nxmin, nxmax
       implicit none
       real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: R, Z
     end subroutine cmp_precond_con
  end interface


  if (verbose) write(*,*) 'Start of conduction_imp'

  E = E - 0.5*sum(rhou**2,4)/rho
  if (allocated(B)) E = E - 0.5*sum(B**2,4)
  call temperature(rho,E,TT,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))
  if(minval(TT(:,:,:))<0.) then
     do k = nxmin(3),nxmax(3)
     do j = nxmin(2),nxmax(2)
     do i = nxmin(1),nxmax(1)
        if(TT(i,j,k)<0.) then
           write(*,101) i,j,k,TT(i,j,k)
101        format('Tloc negative (i,j,k), T=',I5,I5,I5,2x,E10.3)
           call temperature(rho(i,j,k),1.e-8*E(i,j,k),xx,1,1,1,1,1,1)
           TT(i,j,k) = xx
        endif
     enddo
     enddo
     enddo
  endif
  E = E + 0.5*sum(rhou**2,4)/rho
  if (allocated(B)) E = E + 0.5*sum(B**2,4)

  ! Conjugate Gradient method
  TTold=TT
  call cg(TT,TTold,icl_con,mat_prod_con,precond_con,cmp_precond_con,tolx_con,nite,dif)
  nitetot_con=nitetot_con+nite

  ! Update gas energy

  var_loc = 0.

  E = E - 0.5*sum(rhou**2,4)/rho
  if (allocated(B)) E = E - 0.5*sum(B**2,4)
  do k = nxmin(3),nxmax(3)
  do j = nxmin(2),nxmax(2)
  do i = nxmin(1),nxmax(1)
     call temperature(rho(i,j,k),E(i,j,k),Temp,1,1,1,1,1,1)
     var_loc = max(var_loc,2.*abs(TT(i,j,k) - Temp)/(TT(i,j,k) + Temp) )
  enddo
  enddo
  enddo  

  call energy(rho,TT,E,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))
  E = E + 0.5*sum(rhou**2,4)/rho
  if (allocated(B)) E = E + 0.5*sum(B**2,4)

#if PARA==1
  info = 0 
  call MPI_ALLREDUCE(var_loc, varmax ,1,MPI_DOUBLE_PRECISION,MPI_MAX,MPI_COMM_WORLD,info)
#else
  varmax = var_loc
#endif

  if (mype == 0) write(*,997) nite,varmax*100.,dif,nitetot_con
997 format('nite =',I4,' varmax =',1pe9.2,'%   err =',E10.2,'    nitetot =',I8)

  ! Do not limit the timestep if convergence is good
  if (nite .le. 20) then
     dt_con_imp = 1.0E9 * an
  else
     dt_con_imp = vardt_con *varrel_con*dt/Max(vardt_con*varmax,varrel_con)
  endif

  if (verbose) write(*,*) 'End of conduction_imp'

  return
  
end subroutine conduction_imp_gc

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine MAT_PROD_CON
!
!> Computes the matrix product A * X.
!<
subroutine Mat_prod_con(T,AT)
  
  use var
  use var_loc
  use geom
  use divers
  use parameters

  implicit none
  
  integer, dimension(3,3) :: kron
  integer                 :: nnx,nny,nnz,nn
  integer                 :: i,j,k,idim,id
  integer                 :: ip1,jp1,kp1,im1,jm1,km1

  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: T,AT,kappa
  real, dimension(:,:,:,:  ), allocatable                                :: k_gradT
  real :: kap,flux_ener_tot,Cv,densite,temp,cal_kappa_con

  if (verbose) write(*,*) 'Start of Mat_prod_con'

  ! Allocation des tableaux
  nnx = nx(1) ; nny = nx(2) ; nnz = nx(3)

  kron = 0
  do idim=1,3
     kron(idim,idim) = 1
  enddo

  nn  = max(nnx,nny,nnz)
  allocate(xloc(-1:nn+3,ndim)) 

  allocate(Tloc   (nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)       ))
  allocate(k_gradT(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:ndim))

                xloc(-1:nnx+3,1) = x(-1:nnx+3,1)   
 if (ndim.gt.1) xloc(-1:nny+3,2) = x(-1:nny+3,2)   
 if (ndim.gt.2) xloc(-1:nnz+3,3) = x(-1:nnz+3,3)   

  ! Compute geometry
  if(verbose) write(*,*) 'computing geometry'
  call cal_geom(nnx,nny,nnz)

  ! Compute P and T
  if(verbose) write(*,*) 'computing P and T'

  E = E - 0.5*sum(rhou**2,4)/rho
  if (allocated(B)) E = E - 0.5*sum(B**2,4)
  call temperature(rho,E,Tloc,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))
  E = E + 0.5*sum(rhou**2,4)/rho
  if (allocated(B)) E = E + 0.5*sum(B**2,4)

  ! Compute kappa
  if(verbose) write(*,*) 'computing kappa'

  do k = nxmin(3),nxmax(3)
  do j = nxmin(2),nxmax(2)
  do i = nxmin(1),nxmax(1)

     Temp         = Tloc(i,j,k)
     Densite      =  rho(i,j,k)
     kappa(i,j,k) =  cal_kappa_con(Densite,Temp)

  enddo
  enddo
  enddo
    
  ! Compute k_gradT
  do idim = 1,ndim
  
     do k = 1,nnz+kron(idim,3)
     do j = 1,nny+kron(idim,2)
     do i = 1,nnx+kron(idim,1)

        im1 = i - kron(1,idim) ; jm1 = j - kron(2,idim) ; km1 = k - kron(3,idim)

        if(idim.eq.1) then
           id  = i
        else if(idim.eq.2) then
           id  = j
        else if(idim.eq.3) then
           id  = k
        endif

        kap = 0.5*(kappa(im1,jm1,km1) + kappa(i,j,k)) 
        
        k_gradT(i,j,k,idim) = kap * (T(i,j,k) - T(im1,jm1,km1)) / (xloc(id,idim)-xloc(id-1,idim))  
        
     enddo
     enddo
     enddo

  enddo


  AT = T
  
  do idim = 1,ndim
  
     do k =1,nnz
     do j =1,nny
     do i =1,nnx

        ip1 = i + kron(1,idim) ; jp1 = j + kron(2,idim) ; kp1 = k + kron(3,idim)

        flux_ener_tot  = k_gradT(ip1,jp1,kp1,idim)*ds(ip1,jp1,kp1,idim) -  k_gradT(i,j,k,idim)*ds(i,j,k,idim)

        call heat_capacity(rho(i,j,k),T(i,j,k),cv,1,1,1,1,1,1)

        AT(i,j,k) = AT(i,j,k) - flux_ener_tot * dt / dv(i,j,k) /Cv
        
     enddo
     enddo
     enddo
     
  enddo


 ! Deallocate local arrays
 ! geom
  deallocate(dx,dxc)
  deallocate(ds,dv)
 ! others
  deallocate(xloc,xcloc)
  deallocate(Tloc,k_gradT)

  if (verbose) write(*,*) 'End of Mat_prod_con'

  return

end subroutine Mat_prod_con

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine CMP_PRECOND_CON
!
!> This routine computes the preconditioned matrix.
!<
subroutine cmp_precond_con(RR,ZZ)

  use var
  use var_loc
  use geom
  use divers
  use parameters

  implicit none

  integer, dimension(3,3) :: kron
  integer                 :: nnx,nny,nnz,nn
  integer                 :: i,j,k,idim,id,iface
  integer                 :: ip,jp,kp,im,jm,km

  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) ::RR,ZZ,kappa
  real :: kap,Cv,densite,temp,cal_kappa_con,norm

  if (verbose) write(*,*) 'Start of com_precond_con'

  ! Allocation des tableaux
  nnx = nx(1) ; nny = nx(2) ; nnz = nx(3)

  kron = 0
  do idim=1,3
     kron(idim,idim) = 1
  enddo

  nn  = max(nnx,nny,nnz)
  allocate(xloc(-1:nn+3,ndim)) 

  allocate(Tloc   (nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)       ))

                xloc(-1:nnx+3,1) = x(-1:nnx+3,1)   
 if (ndim.gt.1) xloc(-1:nny+3,2) = x(-1:nny+3,2)   
 if (ndim.gt.2) xloc(-1:nnz+3,3) = x(-1:nnz+3,3)   

  ! Compute geometry
  if(verbose) write(*,*) 'computing geometry'
  call cal_geom(nnx,nny,nnz)

  ! Compute P and T
  if(verbose) write(*,*) 'computing P and T'

  E = E - 0.5*sum(rhou**2,4)/rho
  if (allocated(B)) E = E - 0.5*sum(B**2,4)
  call temperature(rho,E,Tloc,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))
  E = E + 0.5*sum(rhou**2,4)/rho
  if (allocated(B)) E = E + 0.5*sum(B**2,4)

  ! Compute kappa
  if(verbose) write(*,*) 'computing kappa'

  do k = nxmin(3),nxmax(3)
  do j = nxmin(2),nxmax(2)
  do i = nxmin(1),nxmax(1)

     Temp         = Tloc(i,j,k)
     Densite      =  rho(i,j,k)
     kappa(i,j,k) =  cal_kappa_con(Densite,Temp)

  enddo
  enddo
  enddo
    
  ! compute the diagonal element of the matrix and normalize by it
  do k = 1,nnz
     do j = 1,nny
        do i = 1,nnx
           call heat_capacity(rho(i,j,k),Tloc(i,j,k),cv,1,1,1,1,1,1)

           ! term of dtE
           norm = 1.

           ! term of div.F
           do iface = 1,2
           do idim = 1,ndim
           
              if (iface.eq.1) then
                 ip = i                ; jp = j                ; kp = k
                 im = i - kron(1,idim) ; jm = j - kron(2,idim) ; km = k - kron(3,idim)
              else
                 ip = i + kron(1,idim) ; jp = j + kron(2,idim) ; kp = k + kron(3,idim)
                 im = i                ; jm = j                ; km = k
              endif

              kap = 0.5*(kappa(im,jm,km) + kappa(ip,jp,kp)) 

              if(idim.eq.1) then
                 id = ip
              else if(idim.eq.2) then
                 id = jp
              else if(idim.eq.3) then
                 id = kp
              endif
              
              norm = norm + kap*rho(i,j,k) / (xloc(id,idim)-xloc(id-1,idim)) * &
                   dt *ds(ip,jp,kp,idim)/dv(i,j,k) /Cv

           enddo
           enddo

           ZZ(i,j,k) = RR(i,j,k) / norm
           
        enddo
     enddo
  enddo

 ! Deallocate local arrays
 ! geom
  deallocate(dx,dxc)
  deallocate(ds,dv)
 ! others
  deallocate(xloc,xcloc)
  deallocate(Tloc)

  if (verbose) write(*,*) 'End of cmp_precond_con'

  return

end subroutine cmp_precond_con
