!> \file
!! Contains subroutines gravity_step(), pt_mass() and
!! cst_gravity()
!!
!! This file contains the routines which compute different
!! types of gravitational potential.
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine GRAVITY_STEP
!
!> This routine calls the potential computation routines 
!! according to gravity_type.
!<
subroutine gravity_step(init,temps_grav,temps_comm)

  use gravity, only : gravity_type,phi,phiold
  use geom   , only : cartesien
  use para   , only : mype

  implicit none
  
#if PARA==1
  include 'mpif.h'
#else
  real     :: MPI_Wtime
  external :: MPI_Wtime
#endif

  integer :: ierr
  logical :: init
  real    :: temps_comm,temps_grav,tr1,tr2
  
  tr1 = MPI_Wtime(ierr)             

  ! Store potential at time n in phiold
  if (init) then
     call allocate_array_gra
  else
     phiold = phi
  endif

  ! Get potential at time n+1
  select case (gravity_type)
  case(0)
     phi = 0.d0
  case(1)
     call gc(init)      ! Self-gravity (Conjugate Gradient)
  case(2)
     call pt_mass ! Point source potential
  case(3)
     if(.not.cartesien) then
        call close_program(mype,'Constant gravitational field but geometry not Cartesian')
     endif
     call cst_gravity ! Constant gravitational field
  end select

  if (init) then
     phiold = phi
  else
     call update_gravity
  endif
  
  tr2 = MPI_Wtime(ierr)     
  temps_grav = temps_grav + tr2-tr1

  call communications_gen(Phi,1,temps_comm)
  call limites_gra(Phi)
  
  call communications(temps_comm)
  call limites

  return
  
end subroutine gravity_step

!###########################################################
!###########################################################
!###########################################################
!###########################################################

! Subroutine PT_MASS
!
!> This routine computes a point-mass gravitational
!! potential.
!!
!! Mass = gravity_params(1)
!!
!! Cartesian geometry: position given by gravity_params(2:4)
!!
!! Cylindrical & Spherical geometry: position at grid center
!<
subroutine pt_mass

  use gravity   , only : gravity_params,phi
  use parameters, only : ndim,nxmin,nxmax
  use geom      , only : x,Cartesien,Cylindrique,Spherique
  use unites    , only : G
  use const
  
  implicit none

  integer :: i,j,k
  real    :: r,mass,xc,yc,zc,xcenter,ycenter,zcenter
  
  mass                 = gravity_params(1)
               xcenter = gravity_params(2) 
  if (ndim>=2) ycenter = gravity_params(3)
  if (ndim==3) zcenter = gravity_params(4)

  if (Cartesien) then
     do k = nxmin(3),nxmax(3)
        zc = half*(x(k+1,3)+x(k,3))
        do j = nxmin(2),nxmax(2)
           yc = half*(x(j+1,2)+x(j,2))
           do i = nxmin(1),nxmax(1)
              xc = half*(x(i+1,1)+x(i,1))
              r = sqrt((xc-xcenter)**2+(yc-ycenter)**2+(zc-zcenter)**2)
              phi(i,j,k) = -G*mass/r
           end do
        end do
     end do
  endif
  if (Cylindrique) then
     do k = nxmin(3),nxmax(3)
        zc = half*(x(k+1,3)+x(k,3))
        do i = nxmin(1),nxmax(1)
           xc = half*(x(i+1,1)+x(i,1))
           r = sqrt(xc**2+zc**2)
           phi(i,nxmin(2):nxmax(2),k)=-G*mass/r
        end do
     end do
  endif
  if (Spherique) then
     do i = nxmin(1),nxmax(1)
        r = half*(x(i+1,1)+x(i,1))
        phi(i,nxmin(2):nxmax(2),nxmin(3):nxmax(3)) = -G*mass/r
     end do
  endif
  
  return

end subroutine pt_mass

!###########################################################
!###########################################################
!###########################################################
!###########################################################

! Subroutine CST_GRAVITY
!
!> This routine computes a potential for a constant
!! gravitational field
!!
!! g = gravity_params(1:ndim)
!<
subroutine cst_gravity

  use gravity   , only : gravity_params,phi
  use parameters, only : ndim,nxmin,nxmax
  use geom      , only : x,Cartesien,Cylindrique,Spherique
  use const
  
  implicit none

  integer :: i,j,k
  real    :: gx,gy,gz,xc,yc,zc

  gy = 0.
  gz = 0.

               gx = gravity_params(1)
  if (ndim>=2) gy = gravity_params(2)
  if (ndim==3) gz = gravity_params(3)
  
  do k = nxmin(3),nxmax(3)
     zc = half*(x(k+1,3)+x(k,3))
     do j = nxmin(2),nxmax(2)
        yc = half*(x(j+1,2)+x(j,2))
        do i = nxmin(1),nxmax(1)
           xc = half*(x(i+1,1)+x(i,1))
           phi(i,j,k) = -(gx*xc+gy*yc+gz*zc)
        end do
     end do
  end do
  
  return

end subroutine cst_gravity
