!> \file
!! Contains module rdwrt_h5 and subroutines
!! dump_1d_array_int_h5(), dump_1d_array_h5(),
!! dump_3d_array_h5(), dump_4d_array_h5(), 
!! dump_1D_array_string_h5(),
!! get_1d_array_int_h5(), get_1d_array_h5(),
!! get_3d_array_h5() and get_4d_array_h5(),
!! get_1D_array_string_h5()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module RDWRT_H5
!
!> Contains reading and writing hdf5 subroutines.
!<
module rdwrt_h5

#if HDF==1
  use hdf5

  contains

!###########################################################
!###########################################################
!###########################################################
!###########################################################

     !  Subroutine DUMP_1D_ARRAY_INT_H5
     !
     !> Writes a 1D array in hdf5 format.
     !<
     subroutine dump_1d_array_int_h5(loc_id,dsetname,array,nx)
     
       use hdf5

       implicit none
   
       integer(hid_t)         :: loc_id
       character(len=*)       :: dsetname
       integer                :: nx
       integer, dimension(nx) :: array
       
       ! HDF5 vars
       integer        :: ierr
       integer(hid_t) :: h5_dspace
       integer(hid_t) :: h5_dset
       
       integer(hsize_t), dimension(1) :: dims
       integer :: rank=1
       
       ! Dump the array in dataset dsetname
       dims = (/ nx /)
       call H5Screate_simple_f(rank, dims, h5_dspace, ierr)
       call H5Dcreate_f(loc_id, trim(dsetname), H5T_NATIVE_INTEGER, &
            & h5_dspace, h5_dset, ierr)
       call H5Dwrite_f(h5_dset, H5T_NATIVE_INTEGER, array, dims, ierr)
       call H5Dclose_f(h5_dset, ierr)
       call H5Sclose_f(h5_dspace, ierr)
       
       return
       
     end subroutine dump_1d_array_int_h5
     
!###########################################################
!###########################################################
!###########################################################
!###########################################################

     !  Subroutine DUMP_1D_ARRAY_H5
     !
     !> Writes a 1D array in hdf5 format.
     !<
     subroutine dump_1d_array_h5(loc_id,dsetname,array,nx)
     
       use hdf5
       
       implicit none
   
       integer(hid_t)        :: loc_id
       character(LEN=*)      :: dsetname
       integer               :: nx
       real*8, dimension(nx) :: array
       
       ! HDF5 vars
       integer        :: ierr
       integer(hid_t) :: h5_dspace
       integer(hid_t) :: h5_dset
       
       integer(hsize_t), dimension(1) :: dims
       integer :: rank=1
       
       ! Dump the array in dataset dsetname
       dims = (/ nx /)
       call H5Screate_simple_f(rank, dims, h5_dspace, ierr)
       call H5Dcreate_f(loc_id, trim(dsetname), H5T_NATIVE_DOUBLE, &
            & h5_dspace, h5_dset, ierr)
       call H5Dwrite_f(h5_dset, H5T_NATIVE_DOUBLE, array, dims, ierr)
       call H5Dclose_f(h5_dset, ierr)
       call H5Sclose_f(h5_dspace, ierr)
       
       return
       
     end subroutine dump_1d_array_h5
     
!###########################################################
!###########################################################
!###########################################################
!###########################################################

     !  Subroutine DUMP_3D_ARRAY_H5
     !
     !> Writes a 3D array in hdf5 format.
     !<
     subroutine dump_3d_array_h5(loc_id,dsetname,array,nx,ny,nz)
     
       use hdf5
     
       implicit none
       
       integer(hid_t)              :: loc_id
       character(LEN=*)            :: dsetname
       integer                     :: nx,ny,nz
       real*8, dimension(nx,ny,nz) :: array
       
       ! HDF5 vars
       integer        :: ierr
       integer(hid_t) :: h5_dspace
       integer(hid_t) :: h5_dset
       
       integer(hsize_t), dimension(3) :: dims
       integer :: rank=3
       
       ! Dump the array in dataset dsetname
       dims = (/ nx,ny,nz /)
       call H5Screate_simple_f(rank, dims, h5_dspace, ierr)
       call H5Dcreate_f(loc_id, trim(dsetname), H5T_NATIVE_DOUBLE, &
            & h5_dspace, h5_dset, ierr)
       call H5Dwrite_f(h5_dset, H5T_NATIVE_DOUBLE, array, dims, ierr)
       call H5Dclose_f(h5_dset, ierr)
       call H5Sclose_f(h5_dspace, ierr)
       
       return
       
     end subroutine dump_3d_array_h5
     
!###########################################################
!###########################################################
!###########################################################
!###########################################################

     !  Subroutine DUMP_4D_ARRAY_H5
     !
     !> Writes a 4D array in hdf5 format.
     !<
     subroutine dump_4d_array_h5(loc_id,dsetname,array,n1,n2,n3,n4)
     
       use hdf5
     
       implicit none
       
       integer(hid_t)                 :: loc_id
       character(LEN=*)               :: dsetname
       integer                        :: n1,n2,n3,n4
       real*8, dimension(n1,n2,n3,n4) :: array
       
       ! HDF5 vars
       integer        :: ierr
       integer(hid_t) :: h5_dspace
       integer(hid_t) :: h5_dset
       
       integer(hsize_t), dimension(4) :: dims
       integer :: rank=4
       
       ! Dump the array in dataset dsetname
       dims = (/ n1,n2,n3,n4 /)
       call H5Screate_simple_f(rank, dims, h5_dspace, ierr)
       call H5Dcreate_f(loc_id, trim(dsetname), H5T_NATIVE_DOUBLE, &
            & h5_dspace, h5_dset, ierr)
       call H5Dwrite_f(h5_dset, H5T_NATIVE_DOUBLE, array, dims, ierr)
       call H5Dclose_f(h5_dset, ierr)
       call H5Sclose_f(h5_dspace, ierr)
       
       return
       
     end subroutine dump_4d_array_h5

!###########################################################
!###########################################################
!###########################################################
!###########################################################

     !  Subroutine DUMP_1D_ARRAY_STRING_H5
     !
     !> Writes a 1D string array in hdf5 format.
     !<
     subroutine dump_1D_array_string_h5(loc_id,dsetname,array,length)
     
       use hdf5
     
       implicit none
       
       integer(hid_t)                      :: loc_id
       integer(hid_t)                      :: type_id
       character(LEN=*)                    :: dsetname
       integer                             :: length
       character(LEN=length)               :: array
       

       ! HDF5 vars
       integer        :: ierr
       integer(hid_t) :: h5_dspace
       integer(hid_t) :: h5_dset
       
       integer(hsize_t), dimension(1) :: dims
       integer :: rank=1
       
       ! Dump the array in dataset dsetname
       dims = (/ 1 /)
       call H5Screate_simple_f(rank, dims, h5_dspace, ierr)


       
       call H5Tcopy_f(H5T_NATIVE_CHARACTER,type_id,ierr)
       call H5Tset_size_f(type_id,length,ierr)

       call H5Dcreate_f(loc_id, trim(dsetname), type_id, &
            & h5_dspace, h5_dset, ierr)
       call H5Dwrite_f(h5_dset, type_id, array, dims, ierr)
       call H5Dclose_f(h5_dset, ierr)
       call H5Sclose_f(h5_dspace, ierr)

       return
       
     end subroutine dump_1D_array_string_h5

!###########################################################
!###########################################################
!###########################################################
!###########################################################

     !  Subroutine GET_1D_ARRAY_INT_H5
     !
     !> Reads a 1D array in hdf5 format.
     !<
     subroutine get_1d_array_int_h5(loc_id,dsetname,array,nx)
     
       use hdf5
     
       implicit none
       
       integer(hid_t)               :: loc_id
       character(LEN=*)             :: dsetname
       integer                      :: nx
       integer, dimension(nx)       :: array
     
       ! HDF5 vars
       integer        :: ierr
       integer(hid_t) :: h5_dset
       
       integer(hsize_t), dimension(1) :: dims
       
       dims=(/ nx /)
       call H5Dopen_f(loc_id,trim(dsetname),h5_dset,ierr)
       call H5Dread_f(h5_dset,H5T_NATIVE_INTEGER,array,dims,ierr)
       call H5Dclose_f(h5_dset,ierr)
       
       return
       
     end subroutine get_1d_array_int_h5

!###########################################################
!###########################################################
!###########################################################
!###########################################################

     !  Subroutine GET_1D_ARRAY_H5
     !
     !> Reads a 1D array in hdf5 format.
     !<
     subroutine get_1d_array_h5(loc_id,dsetname,array,nx)
     
       use hdf5
     
       implicit none
       
       integer(hid_t)              :: loc_id
       character(LEN=*)            :: dsetname
       integer                     :: nx
       real*8, dimension(nx)       :: array
     
       ! HDF5 vars
       integer        :: ierr
       integer(hid_t) :: h5_dset
       
       integer(hsize_t), dimension(1) :: dims
       
       dims=(/ nx /)
       call H5Dopen_f(loc_id,trim(dsetname),h5_dset,ierr)
       call H5Dread_f(h5_dset,H5T_NATIVE_DOUBLE,array,dims,ierr)
       call H5Dclose_f(h5_dset,ierr)
       
       return
       
     end subroutine get_1d_array_h5

!###########################################################
!###########################################################
!###########################################################
!###########################################################

     !  Subroutine GET_3D_ARRAY_H5
     !
     !> Reads a 3D array in hdf5 format.
     !<
     subroutine get_3d_array_h5(loc_id,dsetname,array,nx,ny,nz)
     
       use hdf5
     
       implicit none
       
       integer(hid_t)              :: loc_id
       character(LEN=*)            :: dsetname
       integer                     :: nx,ny,nz
       real*8, dimension(nx,ny,nz) :: array
     
       ! HDF5 vars
       integer        :: ierr
       integer(hid_t) :: h5_dset
       
       integer(hsize_t), dimension(3) :: dims
       
       dims=(/ nx,ny,nz /)
       call H5Dopen_f(loc_id,trim(dsetname),h5_dset,ierr)
       call H5Dread_f(h5_dset,H5T_NATIVE_DOUBLE,array,dims,ierr)
       call H5Dclose_f(h5_dset,ierr)
       
       return
     
     end subroutine get_3d_array_h5
     
!###########################################################
!###########################################################
!###########################################################
!###########################################################

     !  Subroutine GET_4D_ARRAY_H5
     !
     !> Reads a 4D array in hdf5 format.
     !<
     subroutine get_4d_array_h5(loc_id,dsetname,array,n1,n2,n3,n4)
     
       use hdf5
       
       implicit none
       
       integer(hid_t)                 :: loc_id
       character(LEN=*)               :: dsetname
       integer                        :: n1,n2,n3,n4
       real*8, dimension(n1,n2,n3,n4) :: array
     
       ! HDF5 vars
       integer        :: ierr
       integer(hid_t) :: h5_dset
       
       integer(hsize_t), dimension(4) :: dims
       
       dims=(/ n1,n2,n3,n4 /)
       call H5Dopen_f(loc_id,trim(dsetname),h5_dset,ierr)
       call H5Dread_f(h5_dset,H5T_NATIVE_DOUBLE,array,dims,ierr)
       call H5Dclose_f(h5_dset,ierr)
       
       return
       
     end subroutine get_4d_array_h5

!###########################################################
!###########################################################
!###########################################################
!###########################################################

     !  Subroutine GET_1D_ARRAY_STRING_H5
     !
     !> Reads a 1D string array in hdf5 format.
     !<
     subroutine get_1D_array_string_h5(loc_id,dsetname,array,length)
     
       use hdf5
     
       implicit none
       
       integer(hid_t)                      :: loc_id
       integer(hid_t)                      :: type_id
       character(LEN=*)                    :: dsetname
       integer                             :: length,n1
       character(LEN=length) :: array
       
       ! HDF5 vars
       integer        :: ierr
       integer(hid_t) :: h5_dset
       
       integer(hsize_t), dimension(1) :: dims
     
       dims = (/ 1 /)
      
       call H5Tcopy_f(H5T_NATIVE_CHARACTER,type_id,ierr)
       call H5Tset_size_f(type_id,length,ierr)

       call H5Dopen_f(loc_id,trim(dsetname),h5_dset,ierr)
       call H5Dread_f(h5_dset,type_id,array,dims,ierr)
       call H5Dclose_f(h5_dset,ierr)

       return
       
     end subroutine get_1D_array_string_h5

#endif

end module rdwrt_h5
