!> \file
!! Contains module prime and subroutines find_prime(),
!! primedec() and cpudec()
!!
!! This file contains the subroutines which are used to
!! decompose the computational volume and split it evenly
!! between the cpus.
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Module PRIME
!
!> Contains variables used for prime number computations.
!<
module prime
  implicit none
  integer, parameter       :: Pmax  = 50000      !< Pmax
  integer, dimension(Pmax) :: PrimeNumber        !< PrimeNumber
  integer, dimension(Pmax) :: Number             !< Number
  integer, parameter       :: Pdmax = 200        !< Pdmax
  integer                  :: Nprime             !< Nprime
end module prime

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine FIND_PRIME
!
!> Decomposes Number into prime factors.
!<
subroutine find_prime

  use prime

  implicit none

  integer :: i,j

  Number = 1
  
  i = 2
  do while (i < Pmax)
     if (Number(i) == 1) then
        do j=2*i,Pmax,i
           Number(j) = 0
        enddo
     endif
     i = i+1
  enddo

  j = 0
  do i=1,Pmax
     if(Number(i) == 1) then
        j=j+1
        PrimeNumber(j) = i
     endif
  enddo
  Nprime = j
  
  return
  
end subroutine find_prime

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine PRIMEDEC
!
!> Decomposes n into prime factors.
!<
subroutine PrimeDec(n,Dec,Np)

  use prime
  use para , only : mype

  implicit none

  integer, dimension(Pdmax) :: Dec
  integer                   :: n,Np
  integer                   :: m,i


  if (n > Pmax) then
     write(*,*) 'PrimeDec was called with n too large'
     write(*,*) 'N    = ',N
     write(*,*) 'Pmax = ',Pmax
     call close_program(mype,'called from PRIMEDEC')
  endif

  Dec = 0

  m  = n
  i  = 2
  Np = 0
  do while(m > 1) 
     if (mod(m,PrimeNumber(i)) == 0 ) then
        Np = Np + 1
        Dec(Np) = PrimeNumber(i)
        m =  m/PrimeNumber(i)
     else
        i = i + 1
     endif
  enddo

  return
  
end subroutine PrimeDec
  
!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine CPUDEC
!
!> Decomposes total number of cpus into prime factors.
!<
subroutine CpuDec(Ncpu,Cpu_dim,NN)

  use parameters , only : Ndim
  use prime      

  implicit none

  integer, dimension(Ndim ) :: Cpu_Dim,NN,Cpu_dimt
  integer, dimension(Pdmax) :: Dec, Combi
  integer                   :: Ncpu
  integer                   :: Ncombi, Np
  integer                   :: i,j,m, dn,dnmin,ppmax,ppmin,pp


  if (Ndim == 1) then
     Cpu_dim(1) = Ncpu
     return
  endif

  call find_prime
  call PrimeDec(Ncpu,Dec,Np)

  Ncombi = Ndim**Np
  dnmin  = 1000000
  
  do i = 0,Ncombi-1
     
     m        = i
     Cpu_dimt = 1
     do j=1,Np
        Combi(j) = m/Ndim**(Np-j)
         m = m - Combi(j)*Ndim**(Np-j)
         Cpu_dimt(Combi(j)+1) =  Cpu_dimt(Combi(j)+1) * Dec(j) 
     enddo
     
     ppmax = 0
     ppmin = 10000000 
     do j=1,Ndim
        pp    = NN(j)/Cpu_dimt(j)
        ppmax = max(pp,ppmax)
        ppmin = min(pp,ppmin)
     enddo
     
     dn = ppmax-ppmin

     if (dn < dnmin) then
        dnmin   = dn
        Cpu_Dim = Cpu_dimt
     endif

  enddo

  return
  
end subroutine CpuDec
