!> \file
!! Contains subroutine slopes() and functions minmod(),
!! moncen(), moyharm() and vavl()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine SLOPES
!
!> Computes slopes for hydro and radiative transfer
!! variables in 3 dimensions.
!! The different slope limiters are:
!! - 1 : Minmod
!! - 2 : Moncen
!! - 3 : Positivity preserving
!! - 4 : Harmonic average
!! - 5 : Van Albada & Van Leer
!! - 6 : Superbee
!<
subroutine slopes(xx,pxxG,pxxD,nxmin,nxmax,nymin,nymax,nzmin,nzmax,slope_type) 

  use parameters, only : ndim
  use geom      , only : dxc,geometrie,xcloc,geom_dir
  use const

  implicit none

  integer :: nxmin,nxmax,nymin,nymax,nzmin,nzmax
  integer :: ilo,ihi,jlo,jhi,klo,khi
  integer :: i,j,k,slope_type

  real, dimension(nxmin:nxmax,nymin:nymax,nzmin:nzmax       ) :: xx
  real, dimension(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:ndim) :: pxxG,pxxD

  real :: dfll,dflm,dflr,dfml,dfmm,dfmr,dfrl,dfrm,dfrr,dfcen
  real :: dflll,dflml,dflrl,dfmll,dfmml,dfmrl,dfrll,dfrml,dfrrl
  real :: dfllm,dflmm,dflrm,dfmlm,dfmmm,dfmrm,dfrlm,dfrmm,dfrrm
  real :: dfllr,dflmr,dflrr,dfmlr,dfmmr,dfmrr,dfrlr,dfrmr,dfrrr
  real :: vmin,vmax,dfx,dfy,dfz,dff
  real :: dlim,slop

  pxxG = 0.0
  pxxD = 0.0
  
  if(slope_type == 0)then
  
     return
     
  else

     ilo = min(1, nxmin+1) ; ihi = max(1, nxmax-1)
     jlo = min(1, nymin+1) ; jhi = max(1, nymax-1)
     klo = min(1, nzmin+1) ; khi = max(1, nzmax-1)

     SELECT CASE(slope_type) ! select slope limiter
     
     CASE(1) ! MINMOD

        do k = klo, khi
           do j = jlo, jhi
              do i = ilo, ihi
	      
                 dflmm = ( xx(i-1,j  ,k  )-xx(i,j,k) )/ dxc(i  ,1)
                 dfrmm = ( xx(i+1,j  ,k  )-xx(i,j,k) )/ dxc(i+1,1)
                 pxxG(i,j,k,1) = MinMod(dfrmm, -dflmm)

                 if(ndim.gt.1) then  ! slopes along y direction
                    dfmlm = ( xx(i  ,j-1,k  )-xx(i,j,k) )/ dxc(j  ,2)
                    dfmrm = ( xx(i  ,j+1,k  )-xx(i,j,k) )/ dxc(j+1,2)
                    pxxG(i,j,k,2) = MinMod(dfmrm, -dfmlm)
                 end if

                 if(ndim.gt.2) then  ! slopes along z direction
                    dfmml = ( xx(i  ,j  ,k-1)-xx(i,j,k) )/ dxc(k  ,3)
                    dfmmr = ( xx(i  ,j  ,k+1)-xx(i,j,k) )/ dxc(k+1,3)
                    pxxG(i,j,k,3) = MinMod(dfmmr, -dfmml)
                 end if

                 pxxD(i,j,k,:) = pxxG(i,j,k,:)

              end do
           end do
        end do

     CASE(2) ! MONCEN

        do k = klo, khi
           do j = jlo, jhi
              do i = ilo, ihi

                 dflmm = two * ( xx(i,  j,  k  )-xx(i-1,j  ,k  ) ) / dxc(i  ,1)
                 dfrmm = two * ( xx(i+1,j  ,k  )-xx(i  ,j  ,k  ) ) / dxc(i+1,1)
                 dfcen = ( xx(i+1,j  ,k  )-xx(i-1,j  ,k  ) ) / ( dxc(i+1,1)+dxc(i,1) )
                 pxxG(i,j,k,1) = moncen(dflmm,dfrmm,dfcen)

                 if(ndim.gt.1) then  ! slopes along y direction
                    dfmlm = two * ( xx(i,  j,  k  )-xx(i  ,j-1,k  ) ) / dxc(j  ,2)
                    dfmrm = two * ( xx(i  ,j+1,k  )-xx(i  ,j  ,k  ) ) / dxc(j+1,2)
                    dfcen = ( xx(i  ,j+1,k  )-xx(i  ,j-1,k  ) ) / ( dxc(j+1,2)+dxc(j,2) )
                    pxxG(i,j,k,2) = moncen(dfmlm,dfmrm,dfcen)
                 end if

                 if(ndim.gt.2) then  ! slopes along z direction
                    dfmml = two * ( xx(i,  j,  k  )-xx(i  ,j  ,k-1) ) / dxc(k  ,3)
                    dfmmr = two * ( xx(i  ,j  ,k+1)-xx(i  ,j  ,k  ) ) / dxc(k+1,3)
                    dfcen = ( xx(i  ,j  ,k+1)-xx(i  ,j  ,k-1) ) / ( dxc(k+1,3)+dxc(k,3) )
                    pxxG(i,j,k,3) = moncen(dfmml,dfmmr,dfcen)
                 end if

                 pxxD(i,j,k,:) = pxxG(i,j,k,:)

              end do
           end do
        end do

     CASE(3) ! positivity preserving

        if( ndim == 1) then
           do k = klo, khi
              do j = jlo, jhi
                 do i = ilo, ihi
                    dflm = xx(i-1,j  ,k)-xx(i,j,k)
                    dfmm = xx(i  ,j  ,k)-xx(i,j,k)
                    dfrm = xx(i+1,j  ,k)-xx(i,j,k)

                    vmin = min(dflm,dfmm,dfrm)
                    vmax = max(dflm,dfmm,dfrm)

                    dfx  = 0.5*(xx(i+1,j,k)-xx(i-1,j,k))
                    dff  = abs(dfx)

                    if(dff>0.0)then
                       slop = min(1.0,min(abs(vmin),abs(vmax))/dff)
                    else
                       slop = 1.0
                    endif

                    dlim = slop

                    pxxG(i,j,k,1) = dlim*dfx/ (dxc(i  ,1) )
                    pxxD(i,j,k,:) = pxxG(i,j,k,:)
                 end do
              end do
           end do

        elseif ( ndim == 2) then
           do k = klo, khi
              do j = jlo, jhi
                 do i = ilo, ihi
                    dfll = xx(i-1,j-1,k)-xx(i,j,k)
                    dflm = xx(i-1,j  ,k)-xx(i,j,k)
                    dflr = xx(i-1,j+1,k)-xx(i,j,k)
                    dfml = xx(i  ,j-1,k)-xx(i,j,k)
                    dfmm = xx(i  ,j  ,k)-xx(i,j,k)
                    dfmr = xx(i  ,j+1,k)-xx(i,j,k)
                    dfrl = xx(i+1,j-1,k)-xx(i,j,k)
                    dfrm = xx(i+1,j  ,k)-xx(i,j,k)
                    dfrr = xx(i+1,j+1,k)-xx(i,j,k)

                    vmin = min(dfll,dflm,dflr,dfml,dfmm,dfmr,dfrl,dfrm,dfrr)
                    vmax = max(dfll,dflm,dflr,dfml,dfmm,dfmr,dfrl,dfrm,dfrr)

                    dfx  = 0.5*(xx(i+1,j,k)-xx(i-1,j,k))
                    dfy  = 0.5*(xx(i,j+1,k)-xx(i,j-1,k))
                    dff  = 0.5*(abs(dfx)+abs(dfy))

                    if(dff>0.0)then
                       slop = min(1.0,min(abs(vmin),abs(vmax))/dff)
                    else
                       slop = 1.0
                    endif

                    dlim = slop

                    pxxG(i,j,k,1) = dlim*dfx/ (dxc(i  ,1) )
                    pxxG(i,j,k,2) = dlim*dfy/ (dxc(j  ,2) )
                    pxxD(i,j,k,:) = pxxG(i,j,k,:)
                 end do
              end do
           end do
        else
           do k = klo, khi
              do j = jlo, jhi
                 do i = ilo, ihi

                    dflll = xx(i-1,j-1,k-1)-xx(i,j,k)
                    dflml = xx(i-1,j  ,k-1)-xx(i,j,k)
                    dflrl = xx(i-1,j+1,k-1)-xx(i,j,k)
                    dfmll = xx(i  ,j-1,k-1)-xx(i,j,k)
                    dfmml = xx(i  ,j  ,k-1)-xx(i,j,k)
                    dfmrl = xx(i  ,j+1,k-1)-xx(i,j,k)
                    dfrll = xx(i+1,j-1,k-1)-xx(i,j,k)
                    dfrml = xx(i+1,j  ,k-1)-xx(i,j,k)
                    dfrrl = xx(i+1,j+1,k-1)-xx(i,j,k)

                    dfllm = xx(i-1,j-1,k  )-xx(i,j,k)
                    dflmm = xx(i-1,j  ,k  )-xx(i,j,k)
                    dflrm = xx(i-1,j+1,k  )-xx(i,j,k)
                    dfmlm = xx(i  ,j-1,k  )-xx(i,j,k)
                    dfmmm = 0.0
                    dfmrm = xx(i  ,j+1,k  )-xx(i,j,k)
                    dfrlm = xx(i+1,j-1,k  )-xx(i,j,k)
                    dfrmm = xx(i+1,j  ,k  )-xx(i,j,k)
                    dfrrm = xx(i+1,j+1,k  )-xx(i,j,k)

                    dfllr = xx(i-1,j-1,k+1)-xx(i,j,k)
                    dflmr = xx(i-1,j  ,k+1)-xx(i,j,k)
                    dflrr = xx(i-1,j+1,k+1)-xx(i,j,k)
                    dfmlr = xx(i  ,j-1,k+1)-xx(i,j,k)
                    dfmmr = xx(i  ,j  ,k+1)-xx(i,j,k)
                    dfmrr = xx(i  ,j+1,k+1)-xx(i,j,k)
                    dfrlr = xx(i+1,j-1,k+1)-xx(i,j,k)
                    dfrmr = xx(i+1,j  ,k+1)-xx(i,j,k)
                    dfrrr = xx(i+1,j+1,k+1)-xx(i,j,k)


                    vmin = min(dflll,dflml,dflrl,dfmll,dfmml,dfmrl,dfrll,dfrml,dfrrl, &
                         &     dfllm,dflmm,dflrm,dfmlm,dfmmm,dfmrm,dfrlm,dfrmm,dfrrm, &
                         &     dfllr,dflmr,dflrr,dfmlr,dfmmr,dfmrr,dfrlr,dfrmr,dfrrr)
                    vmax = max(dflll,dflml,dflrl,dfmll,dfmml,dfmrl,dfrll,dfrml,dfrrl, &
                         &     dfllm,dflmm,dflrm,dfmlm,dfmmm,dfmrm,dfrlm,dfrmm,dfrrm, &
                         &     dfllr,dflmr,dflrr,dfmlr,dfmmr,dfmrr,dfrlr,dfrmr,dfrrr)

                    dfx  = 0.5*(xx(i+1,j,k)-xx(i-1,j,k))
                    dfy  = 0.5*(xx(i,j+1,k)-xx(i,j-1,k))
                    dfz  = 0.5*(xx(i,j,k+1)-xx(i,j,k-1))
                    dff  = 0.5*(abs(dfx)+abs(dfy)+abs(dfz))

                    if(dff>0.0)then
                       slop = min(1.0,min(abs(vmin),abs(vmax))/dff)
                    else
                       slop = 1.0
                    endif

                    dlim = slop

                    pxxG(i,j,k,1) = dlim*dfx/ (dxc(i  ,1) )
                    if(ndim.gt.1) then  ! slopes along y direction
                       pxxG(i,j,k,2) = dlim*dfy/ (dxc(j  ,2) )
                    end if
                    if(ndim.gt.1) then  ! slopes along z direction
                       pxxG(i,j,k,3) = dlim*dfz/ (dxc(k  ,3) )
                    end if

                    pxxD(i,j,k,:) = pxxG(i,j,k,:)

                 end do
              end do
           end do
        end if

     CASE(4) ! HARMONIC AVERAGE

        do k = klo, khi
           do j = jlo, jhi
              do i = ilo, ihi

                 dflmm = ( xx(i-1,j  ,k  )-xx(i,j,k) )/ dxc(i  ,1)
                 dfrmm = ( xx(i+1,j  ,k  )-xx(i,j,k) )/ dxc(i+1,1)
                 pxxG(i,j,k,1) = MoyHarm(dfrmm, -dflmm)

                 if(ndim.gt.1) then  ! slopes along y direction
                    dfmlm = ( xx(i  ,j-1,k  )-xx(i,j,k) )/ dxc(j  ,2)
                    dfmrm = ( xx(i  ,j+1,k  )-xx(i,j,k) )/ dxc(j+1,2)
                    pxxG(i,j,k,2) = MoyHarm(dfmrm, -dfmlm)
                 end if

                 if(ndim.gt.2) then  ! slopes along z direction
                    dfmml = ( xx(i  ,j  ,k-1)-xx(i,j,k) )/ dxc(k  ,3)
                    dfmmr = ( xx(i  ,j  ,k+1)-xx(i,j,k) )/ dxc(k+1,3)
                    pxxG(i,j,k,3) = MoyHarm(dfmmr, -dfmml)
                 end if

                 pxxD(i,j,k,:) = pxxG(i,j,k,:)

              end do
           end do
        end do

     CASE(5) ! Van Albada Van Leer

        do k = klo, khi
           do j = jlo, jhi
              do i = ilo, ihi

                 dflmm = ( xx(i-1,j  ,k  )-xx(i,j,k) )/ dxc(i  ,1)
                 dfrmm = ( xx(i+1,j  ,k  )-xx(i,j,k) )/ dxc(i+1,1)
                 pxxG(i,j,k,1) = VAVL(dfrmm, -dflmm )

                 if(ndim.gt.1) then  ! slopes along y direction
                    dfmlm = ( xx(i  ,j-1,k  )-xx(i,j,k) )/ dxc(j  ,2)
                    dfmrm = ( xx(i  ,j+1,k  )-xx(i,j,k) )/ dxc(j+1,2)
                    pxxG(i,j,k,2) = VAVL(dfmrm, -dfmlm)
                 end if

                 if(ndim.gt.2) then  ! slopes along z direction
                    dfmml = ( xx(i  ,j  ,k-1)-xx(i,j,k) )/ dxc(k  ,3)
                    dfmmr = ( xx(i  ,j  ,k+1)-xx(i,j,k) )/ dxc(k+1,3)
                    pxxG(i,j,k,3) = VAVL(dfmmr, -dfmml)
                 end if

                 pxxD(i,j,k,:) = pxxG(i,j,k,:)

              end do
           end do
        end do

     CASE(6:) ! Superbee

        do k = klo, khi
           do j = jlo, jhi
              do i = ilo, ihi

                 dflmm = ( xx(i-1,j  ,k  )-xx(i,j,k) )/ dxc(i  ,1)
                 dfrmm = ( xx(i+1,j  ,k  )-xx(i,j,k) )/ dxc(i+1,1)
                 pxxD(i,j,k,1) = MinMod(dfrmm, -2.0*dflmm)
                 pxxG(i,j,k,1) = MinMod(2.0*dfrmm, -dflmm)

                 if(ndim.gt.1) then  ! slopes along y direction
                    dfmlm = ( xx(i  ,j-1,k  )-xx(i,j,k) )/ dxc(j  ,2)
                    dfmrm = ( xx(i  ,j+1,k  )-xx(i,j,k) )/ dxc(j+1,2)
                    pxxD(i,j,k,2) = MinMod(dfmrm, -2.0*dfmlm)
                    pxxG(i,j,k,2) = MinMod(2.0*dfmrm, -dfmlm)
                 end if

                 if(ndim.gt.2) then  ! slopes along z direction
                    dfmml = ( xx(i  ,j  ,k-1)-xx(i,j,k) )/ dxc(k  ,3)
                    dfmmr = ( xx(i  ,j  ,k+1)-xx(i,j,k) )/ dxc(k+1,3)
                    pxxD(i,j,k,3) = MinMod(dfmmr, -2.0*dfmml)
                    pxxG(i,j,k,3) = MinMod(2.0*dfmmr, -dfmml)
                 end if

              end do
           end do
        end do

     END SELECT

  END if

CONTAINS

!###########################################################
!###########################################################
!###########################################################
!###########################################################

  !  Function MINMOD
  !
  !> Minmod averaging function.
  !<
  function minmod(pa,pb) result(mmod)
    
    implicit none
    
    real, intent(in) :: pa,pb
    real             :: mmod

    MMod = 0.5*MIN( ABS(pa), ABS(pb) )*( SIGN(1.0,pa) + SIGN(1.0,pb) )

    return

  end function minmod

!###########################################################
!###########################################################
!###########################################################
!###########################################################

  !  Function MONCEN
  !
  !> Monotonised centred averaging function.
  !<
  function moncen(pa,pb,pc) result(mcen)
  
    implicit none
    
    real, intent(in) :: pa,pb,pc
    real             :: mcen
  
    if((abs(pa) < abs(pb)).and.(abs(pa) < abs(pc)))then
       mcen = pa
    elseif((abs(pb) < abs(pa)).and.(abs(pb) < abs(pc)))then
       mcen = pb
    else
       mcen = pc
    endif

    if((pa*pb < 0.0).or.(pa*pc < 0.0).or.(pb*pc < 0.0))then
       mcen = 0.0
    endif
    
    return
   
  end function moncen

!###########################################################
!###########################################################
!###########################################################
!###########################################################

  !  Function MOYHARM
  !
  !> Harmonic averaging function.
  !<
  function moyharm(pa,pb) result(mmod)
  
    implicit none

    real, intent(in) :: pa,pb
    real             :: mmod

    if (pa*pb .gt. 0.) then
       MMOD = 2.*pa*pb/(pa+pb)
    else
       MMod = 0.
    endif

    return
    
  end function moyharm

!###########################################################
!###########################################################
!###########################################################
!###########################################################

  !  Function VAVL
  !
  !> Van Albada & Van Leer averaging function.
  !<
  function vavl(pa,pb) result(vv)
  
    implicit none
  
    real, intent(in):: pa,pb
    real :: vv,e,num,den
    
    e = 1.0d-08
    
    if(pa*pb > 0.0)then
       num = ( pa*pa + e*e )*pb + ( pb*pb + e*e )*pa
       den = ( pa*pa +  pb*pb + 2.0*e*e )
       vv  = num/den
    else
       vv  = 0.0
    end if
    
    return

  end function vavl

end subroutine slopes
