!> \file
!! contains subroutines cmpflxm(), cmp_mag_flx(),
!! fast_mhd_speed() and uslope()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine CMPFLXM
!
!> UNSPLIT      Unsplit first (second) order Godunov integrator for
!!              polytropic magnetized gas dynamics using 
!!              the Lax-Friedrish sheme
!!              The mesh for the magnetic field is staggered 
!!              The sheme follows closely the paper by
!!              Londrillo & Del Zanna ApJ 2000, 530, 508, 
!!
!!  Inputs/Outputs
!!  uin         => (const)  input state
!! \f$ u_{\mathrm{in}} = (\rho, \rho v_{x}, \rho v_{y}, \rho v_{z}, E_{\mathrm{tot}}, B_{x}, B_{y}, B_{z} )\f$
!!       the hydro variable are centered whereas \f$B_{x}, B_{y}, B_{z}\f$ are written 
!!       on the face (staggered mesh)
!!       note that in MHD one can be in 1 or 2D and have 3 components for v and B
!!       therefore one has 2 variables for the dimension : "ndim" the spatial dimension
!!       and "n_vit" the number of velocity components
!!       
!!  gravin      => (const)  input gravitational acceleration
!!  iu1,iu2     => (const)  first and last index of input array,
!!  ju1,ju2     => (const)  cell centered,    
!!  ku1,ku2     => (const)  including buffer cells.
!!  flux       <=  (modify) return fluxes in the 3 coord directions
!!  if1,if2     => (const)  first and last index of output array,
!!  jf1,jf2     => (const)  edge centered,
!!  kf1,kf2     => (const)  for active cells only.
!!  dx,dy,dz    => (const)  \f$(dx,dy,dz)\f$
!!  dt          => (const)  time step
!!  ngrid       => (const)  number of sub-grids
!!  ndim        => (const)  number of dimensions
!!  n_vit       => (const)  number of velocity components
!<
subroutine cmpflxm(qm,im1,im2,jm1,jm2,km1,km2, &
     &             qp,ip1,ip2,jp1,jp2,kp1,kp2, &
     &                ilo,ihi,jlo,jhi,klo,khi,ln,lt1,lt2,bn,bt1,bt2, &
     &             flx,flx_pre_tot,vel_star,nvar,ndim)
     
  use parameters, only : riemann, n_vit
  use const
  use geom

  implicit none

  integer :: nvar,ndim
  integer :: ln,lt1,lt2,bn,bt1,bt2
  integer :: im1,im2,jm1,jm2,km1,km2
  integer :: ip1,ip2,jp1,jp2,kp1,kp2
  integer :: ilo,ihi,jlo,jhi,klo,khi
  
  real, dimension(im1:im2,jm1:jm2,km1:km2,1:nvar,1:ndim) :: qm
  real, dimension(ip1:ip2,jp1:jp2,kp1:kp2,1:nvar,1:ndim) :: qp 
  real, dimension(ip1:ip2,jp1:jp2,kp1:kp2,1:nvar       ) :: flx
  real, dimension(ip1:ip2,jp1:jp2,kp1:kp2              ) :: flx_pre_tot
  real, dimension(ip1:ip2,jp1:jp2,kp1:kp2,1:ndim,1:ndim) :: vel_star

  ! local variables
  integer :: i,j,k,n,xdim
  
  real, dimension(:,:,:), allocatable :: rS,uS,vS,wS,AS,BS,CS,ptotS  
  real, dimension(:    ), allocatable :: qleft,qright,qgdnv
  real, dimension(:    ), allocatable :: fgdnv,umean
  real :: bn_mean,zero_flux
  real :: ro,uo,vo,wo,bo,co,Ptoto,pressure,cotanxc

  allocate(qleft(nvar),qright(nvar),qgdnv(nvar))
  allocate(fgdnv(nvar),umean(nvar)             )

  allocate(rS   (ip1:ip2,jp1:jp2,kp1:kp2))
  allocate(uS   (ip1:ip2,jp1:jp2,kp1:kp2))
  allocate(vS   (ip1:ip2,jp1:jp2,kp1:kp2))
  allocate(wS   (ip1:ip2,jp1:jp2,kp1:kp2))
  allocate(AS   (ip1:ip2,jp1:jp2,kp1:kp2))
  allocate(BS   (ip1:ip2,jp1:jp2,kp1:kp2))
  allocate(CS   (ip1:ip2,jp1:jp2,kp1:kp2))
  allocate(ptotS(ip1:ip2,jp1:jp2,kp1:kp2))

  xdim=ln-1

  do k = klo, khi
     do j = jlo, jhi
        do i = ilo, ihi
           
           ! Enforce continuity for normal magnetic field
           bn_mean = half*(qm(i,j,k,bn,xdim)+qp(i,j,k,bn,xdim))
           
           ! Left state
           qleft (1) = qm(i,j,k,1  ,xdim) ! Mass density
           qleft (2) = qm(i,j,k,5  ,xdim) ! Pressure
           qleft (3) = qm(i,j,k,ln ,xdim) ! Normal velocity
           qleft (4) = bn_mean            ! Normal magnetic field
           qleft (5) = qm(i,j,k,lt1,xdim) ! Tangential velocity 1
           qleft (6) = qm(i,j,k,bt1,xdim) ! Tangential magnetic field 1
           qleft (7) = qm(i,j,k,lt2,xdim) ! Tangential velocity 2
           qleft (8) = qm(i,j,k,bt2,xdim) ! Tangential magnetic field 2
           
           ! Right state
           qright(1) = qp(i,j,k,1  ,xdim) ! Mass density
           qright(2) = qp(i,j,k,5  ,xdim) ! Pressure
           qright(3) = qp(i,j,k,ln ,xdim) ! Normal velocity
           qright(4) = bn_mean              ! Normal magnetic field
           qright(5) = qp(i,j,k,lt1,xdim) ! Tangential velocity 1
           qright(6) = qp(i,j,k,bt1,xdim) ! Tangential magnetic field 1
           qright(7) = qp(i,j,k,lt2,xdim) ! Tangential velocity 2
           qright(8) = qp(i,j,k,bt2,xdim) ! Tangential magnetic field 2
           
           ! Other advected quantities
           do n = 9, nvar
              qleft (n) = qm(i,j,k,n,xdim)
              qright(n) = qp(i,j,k,n,xdim)    
           end do
           
           ! Solve 1D Riemann problem
           zero_flux = one
           select case (riemann)
           case ('roe')
              call athena_roe    (qleft,qright,fgdnv,zero_flux)
           case ('llf')
              call lax_friedrich (qleft,qright,fgdnv,zero_flux)
           case ('hll')
              call hll           (qleft,qright,fgdnv          )
           case ('hlld')
              call hlld          (qleft,qright,fgdnv          ,ro,uo,vo,wo,bo,co,Ptoto,nvar)
           case ('upwind')
              call lax_friedrich (qleft,qright,fgdnv,zero_flux)
           case ('hydro')
              call hydro_acoustic(qleft,qright,fgdnv          )
           !CASE ('exact')
             !CALL exact         (qleft,qright,fgdnv          )
           end select

           !Swap the variables in order to match the uin convention
           flx(i,j,k,1)      = fgdnv(1)*ds(i,j,k,xdim) !density
           flx(i,j,k,n_vit+2)= fgdnv(2)*ds(i,j,k,xdim) !total energy
           flx(i,j,k,ln)     = fgdnv(3)*ds(i,j,k,xdim) !normal velocity
           flx(i,j,k,bn)     = fgdnv(4)*ds(i,j,k,xdim) !normal magnetic field
           flx(i,j,k,lt1)    = fgdnv(5)*ds(i,j,k,xdim) !transverse velocities 1
           flx(i,j,k,bt1)    = fgdnv(6)*ds(i,j,k,xdim) !transverse magnetic field 1
           flx(i,j,k,lt2)    = fgdnv(7)*ds(i,j,k,xdim) !transverse velocities 2
           flx(i,j,k,bt2)    = fgdnv(8)*ds(i,j,k,xdim) !transverse magnetic field 2

           !Store star state
           rS(i,j,k)    = ro
           uS(i,j,k)    = uo
           vS(i,j,k)    = vo
           wS(i,j,k)    = wo
           AS(i,j,k)    = bn_mean
           BS(i,j,k)    = bo
           CS(i,j,k)    = co
           ptotS(i,j,k) = Ptoto

           !Gather star velocities
           if (xdim==1) then
                          vel_star(i,j,k,1,xdim)=uS(i,j,k)
              if (ndim>1) vel_star(i,j,k,2,xdim)=vS(i,j,k)
              if (ndim>2) vel_star(i,j,k,3,xdim)=wS(i,j,k)
           endif
           if (xdim==2) then
                          vel_star(i,j,k,1,xdim)=wS(i,j,k)
                          vel_star(i,j,k,2,xdim)=uS(i,j,k)
              if (ndim>2) vel_star(i,j,k,3,xdim)=vS(i,j,k)
           endif
           if (xdim==3) then
              vel_star(i,j,k,1,xdim)=wS(i,j,k)
              vel_star(i,j,k,2,xdim)=vS(i,j,k)
              vel_star(i,j,k,3,xdim)=uS(i,j,k)
           endif
           
        end do
     end do
  end do

  ! Add geometrical terms
  if (xdim==1) then
     if (Cartesien) then
        do k = klo, khi
           do j = jlo, jhi
              do i = ilo, ihi-1
                 flx_pre_tot(i,j,k)=-(ptotS(i+1,j,k)*ds(i+1,j,k,1)-ptotS(i,j,k)*ds(i,j,k,1))
              end do
           end do
        end do
     endif
     if (Cylindrique) then
        do k = klo, khi
           do j = jlo, jhi
              do i = ilo, ihi-1              
                 flx_pre_tot(i,j,k)=-dv(i,j,k)*(ptotS(i+1,j,k)-ptotS(i,j,k))/dx(i,1)
                 ro=half*(rS(i,j,k)+rS(i+1,j,k))
                 vo=half*(vS(i,j,k)+vS(i+1,j,k))
                 bo=half*(BS(i,j,k)+BS(i+1,j,k))
                 pressure=(ro*vo*vo-bo*bo)*(ds(i+1,j,k,1)-ds(i,j,k,1))
                 flx_pre_tot(i,j,k)=flx_pre_tot(i,j,k) + pressure
              end do
           end do
        end do
     endif
     if (Spherique) then
        do k = klo, khi
           do j = jlo, jhi
              do i = ilo, ihi-1              
                 flx_pre_tot(i,j,k)=-dv(i,j,k)*(ptotS(i+1,j,k)-ptotS(i,j,k))/dx(i,1)
                 ro=half*(rS(i,j,k)+rS(i+1,j,k))
                 vo=half*(vS(i,j,k)+vS(i+1,j,k))
                 wo=half*(wS(i,j,k)+wS(i+1,j,k))
                 bo=half*(BS(i,j,k)+BS(i+1,j,k))
                 co=half*(CS(i,j,k)+CS(i+1,j,k))
                 pressure=(ro*(vo*vo+wo*wo)-bo*bo-co*co)*(ds(i+1,j,k,1)-ds(i,j,k,1))
                 flx_pre_tot(i,j,k)=flx_pre_tot(i,j,k) + pressure
              end do
           end do
        end do
     endif
  endif

  if (xdim==2) then
     if (Cartesien.or.Cylindrique) then
        do k = klo, khi
           do j = jlo, jhi-1
              do i = ilo, ihi
                 flx_pre_tot(i,j,k)=-(ptotS(i,j+1,k)*ds(i,j+1,k,2)-ptotS(i,j,k)*ds(i,j,k,2))
              end do
           end do
        end do
     endif
     if (Spherique) then
        do k = klo, khi
           do j = jlo, jhi-1
              cotanxc=cos(xcloc(j,2))/sin(xcloc(j,2))
              do i = ilo, ihi
                 flx_pre_tot(i,j,k)=-dv(i,j,k)*(ptotS(i,j+1,k)-ptotS(i,j,k))/dx(j,2)/xcloc(i,1)
                 ro=half*(rS(i,j,k)+rS(i,j+1,k))
                 wo=half*(wS(i,j,k)+wS(i,j+1,k))
                 co=half*(cS(i,j,k)+cS(i,j+1,k))
                 pressure=(ro*wo*wo-co*co)*cotanxc*half*(ds(i+1,j,k,1)-ds(i,j,k,1))
                 flx_pre_tot(i,j,k)=flx_pre_tot(i,j,k) + pressure
              end do
           end do
        end do
     endif
  endif
           
  if (xdim==3) then
     do k = klo, khi-1
        do j = jlo, jhi
           do i = ilo, ihi
              flx_pre_tot(i,j,k)=-(ptotS(i,j,k+1)*ds(i,j,k+1,3)-ptotS(i,j,k)*ds(i,j,k,3))
           end do
        end do
     end do
  endif
           
  deallocate(qleft,qright,qgdnv,fgdnv,umean)
  deallocate(rS,uS,vS,wS,AS,BS,CS,ptotS)
  
  return

end subroutine cmpflxm

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine CMP_MAG_FLX
!
!> 2D Riemann solver to compute EMF at cell edges.
!! Indices of the 2 planar velocity lp1 and lp2 and the
!! orthogonal one, lor and idem for the magnetic field.
!<
subroutine cmp_mag_flx(qRT,irt1,irt2,jrt1,jrt2,krt1,krt2, &
       &               qRB,irb1,irb2,jrb1,jrb2,krb1,krb2, &
       &               qLT,ilt1,ilt2,jlt1,jlt2,klt1,klt2, &
       &               qLB,ilb1,ilb2,jlb1,jlb2,klb1,klb2, &
       &                   ilo ,ihi ,jlo ,jhi ,klo ,khi , &
       &                   lp1 ,lp2 ,lor ,bp1 ,bp2 ,bor , &
       &               emf,nvar,ndim)
  
  use parameters , only : riemann2d, smallc
  use const
  
  implicit none

  integer :: nvar,ndim
  integer :: lp1,lp2,lor,bp1,bp2,bor
  integer :: irt1,irt2,jrt1,jrt2,krt1,krt2
  integer :: irb1,irb2,jrb1,jrb2,krb1,krb2
  integer :: ilt1,ilt2,jlt1,jlt2,klt1,klt2
  integer :: ilb1,ilb2,jlb1,jlb2,klb1,klb2
  integer :: ilo,ihi,jlo,jhi,klo,khi
  
  real, dimension(irt1:irt2,jrt1:jrt2,krt1:krt2,1:nvar,1:3) :: qRT
  real, dimension(irb1:irb2,jrb1:jrb2,krb1:krb2,1:nvar,1:3) :: qRB
  real, dimension(ilt1:ilt2,jlt1:jlt2,klt1:klt2,1:nvar,1:3) :: qLT
  real, dimension(ilb1:ilb2,jlb1:jlb2,klb1:klb2,1:nvar,1:3) :: qLB
  real, dimension(ilb1:ilb2,jlb1:jlb2,klb1:klb2           ) :: emf

  ! local variables
  integer ::i,j,k,xdim
  real, dimension(:), allocatable :: qLL,qRL,qLR,qRR
  real, dimension(:), allocatable :: qleft,qright,fmean_x,fmean_y,qtmp
  real :: ELL,ERL,ELR,ERR,SL,SR,SB,ST,SAL,SAR,SAT,SAB
  real :: zero_flux,E
  real :: cLLx,cRLx,cLRx,cRRx,cLLy,cRLy,cLRy,cRRy
  real :: cfastLLx,cfastRLx,cfastLRx,cfastRRx,cfastLLy,cfastRLy,cfastLRy,cfastRRy
  real :: calfvenR,calfvenL,calfvenT,calfvenB
  real :: vLLx,vRLx,vLRx,vRRx,vLLy,vRLy,vLRy,vRRy
  real :: rLL,rLR,rRL,rRR,pLL,pLR,pRL,pRR,uLL,uLR,uRL,uRR,vLL,vLR,vRL,vRR
  real :: ALL,ALR,ARL,ARR,BLL,BLR,BRL,BRR,CLL,CLR,CRL,CRR
  real :: PtotLL,PtotLR,PtotRL,PtotRR,rcLLx,rcLRx,rcRLx,rcRRx,rcLLy,rcLRy,rcRLy,rcRRy
  real :: ustar,vstar,rstarLLx,rstarLRx,rstarRLx,rstarRRx,rstarLLy,rstarLRy,rstarRLy,rstarRRy
  real :: rstarLL,rstarLR,rstarRL,rstarRR,AstarLL,AstarLR,AstarRL,AstarRR,BstarLL,BstarLR,BstarRL,BstarRR
  real :: EstarLLx,EstarLRx,EstarRLx,EstarRRx,EstarLLy,EstarLRy,EstarRLy,EstarRRy,EstarLL,EstarLR,EstarRL,EstarRR
  real :: AstarT,AstarB,BstarR,BstarL

  allocate(qLL(nvar),qRL(nvar),qLR(nvar),qRR(nvar))
  allocate(qleft(nvar),qright(nvar),fmean_x(nvar),fmean_y(nvar),qtmp(nvar))

  xdim = lor - 1

  do k = klo, khi
     do j = jlo, jhi
        do i = ilo, ihi

           ! Density
           qLL(1) = qRT(i,j,k,1,xdim)
           qRL(1) = qLT(i,j,k,1,xdim)
           qLR(1) = qRB(i,j,k,1,xdim)
           qRR(1) = qLB(i,j,k,1,xdim)

           ! Pressure 
#ifdef ISO
           qLL(2) = qLL(1)*ciso**2
           qRL(2) = qRL(1)*ciso**2
           qLR(2) = qLR(1)*ciso**2
           qRR(2) = qRR(1)*ciso**2
#else
           qLL(2) = qRT(i,j,k,5,xdim)
           qRL(2) = qLT(i,j,k,5,xdim)
           qLR(2) = qRB(i,j,k,5,xdim)
           qRR(2) = qLB(i,j,k,5,xdim)
#endif

           ! First parallel velocity 
           qLL(3) = qRT(i,j,k,lp1,xdim)
           qRL(3) = qLT(i,j,k,lp1,xdim)
           qLR(3) = qRB(i,j,k,lp1,xdim)
           qRR(3) = qLB(i,j,k,lp1,xdim)

           ! Second parallel velocity 
           qLL(4) = qRT(i,j,k,lp2,xdim)
           qRL(4) = qLT(i,j,k,lp2,xdim)
           qLR(4) = qRB(i,j,k,lp2,xdim)
           qRR(4) = qLB(i,j,k,lp2,xdim)

           ! First parallel magnetic field (enforce continuity)
           qLL(6) = half*(qRT(i,j,k,bp1,xdim)+qLT(i,j,k,bp1,xdim))
           qRL(6) = half*(qRT(i,j,k,bp1,xdim)+qLT(i,j,k,bp1,xdim))
           qLR(6) = half*(qRB(i,j,k,bp1,xdim)+qLB(i,j,k,bp1,xdim))
           qRR(6) = half*(qRB(i,j,k,bp1,xdim)+qLB(i,j,k,bp1,xdim))
           
           ! Second parallel magnetic field (enforce continuity)
           qLL(7) = half*(qRT(i,j,k,bp2,xdim)+qRB(i,j,k,bp2,xdim))
           qRL(7) = half*(qLT(i,j,k,bp2,xdim)+qLB(i,j,k,bp2,xdim))
           qLR(7) = half*(qRT(i,j,k,bp2,xdim)+qRB(i,j,k,bp2,xdim))
           qRR(7) = half*(qLT(i,j,k,bp2,xdim)+qLB(i,j,k,bp2,xdim))
           
           ! Orthogonal velocity 
           qLL(5) = qRT(i,j,k,lor,xdim)
           qRL(5) = qLT(i,j,k,lor,xdim)
           qLR(5) = qRB(i,j,k,lor,xdim)
           qRR(5) = qLB(i,j,k,lor,xdim)
           
           ! Orthogonal magnetic Field
           qLL(8) = qRT(i,j,k,bor,xdim)
           qRL(8) = qLT(i,j,k,bor,xdim)
           qLR(8) = qRB(i,j,k,bor,xdim)
           qRR(8) = qLB(i,j,k,bor,xdim)
           
           ! Compute final fluxes
           
           ! vx*by - vy*bx at the four edge centers
           ELL = qLL(3)*qLL(7) - qLL(4)*qLL(6)
           ERL = qRL(3)*qRL(7) - qRL(4)*qRL(6)
           ELR = qLR(3)*qLR(7) - qLR(4)*qLR(6)
           ERR = qRR(3)*qRR(7) - qRR(4)*qRR(6)
           
           if(riemann2d=='hlld')then
              
              rLL=qLL(1); pLL=qLL(2); uLL=qLL(3); vLL=qLL(4); ALL=qLL(6); BLL=qLL(7) ; CLL=qLL(8) 
              rLR=qLR(1); pLR=qLR(2); uLR=qLR(3); vLR=qLR(4); ALR=qLR(6); BLR=qLR(7) ; CLR=qLR(8) 
              rRL=qRL(1); pRL=qRL(2); uRL=qRL(3); vRL=qRL(4); ARL=qRL(6); BRL=qRL(7) ; CRL=qRL(8) 
              rRR=qRR(1); pRR=qRR(2); uRR=qRR(3); vRR=qRR(4); ARR=qRR(6); BRR=qRR(7) ; CRR=qRR(8) 

              ! Compute 4 fast magnetosonic velocity relative to x direction
              qtmp(1)=qLL(1); qtmp(2)=qLL(2); qtmp(7)=qLL(5); qtmp(8)=qLL(8)
              qtmp(3)=qLL(3); qtmp(4)=qLL(6); qtmp(5)=qLL(4); qtmp(6)=qLL(7)
              call find_speed_fast(qtmp,cfastLLx,nvar)
              qtmp(1)=qLR(1); qtmp(2)=qLR(2); qtmp(7)=qLR(5); qtmp(8)=qLR(8)
              qtmp(3)=qLR(3); qtmp(4)=qLR(6); qtmp(5)=qLR(4); qtmp(6)=qLR(7)
              call find_speed_fast(qtmp,cfastLRx,nvar)
              qtmp(1)=qRL(1); qtmp(2)=qRL(2); qtmp(7)=qRL(5); qtmp(8)=qRL(8)
              qtmp(3)=qRL(3); qtmp(4)=qRL(6); qtmp(5)=qRL(4); qtmp(6)=qRL(7)
              call find_speed_fast(qtmp,cfastRLx,nvar)
              qtmp(1)=qRR(1); qtmp(2)=qRR(2); qtmp(7)=qRR(5); qtmp(8)=qRR(8)
              qtmp(3)=qRR(3); qtmp(4)=qRR(6); qtmp(5)=qRR(4); qtmp(6)=qRR(7)
              call find_speed_fast(qtmp,cfastRRx,nvar)
              
              ! Compute 4 fast magnetosonic velocity relative to y direction
              qtmp(1)=qLL(1); qtmp(2)=qLL(2); qtmp(7)=qLL(5); qtmp(8)=qLL(8)
              qtmp(3)=qLL(4); qtmp(4)=qLL(7); qtmp(5)=qLL(3); qtmp(6)=qLL(6)
              call find_speed_fast(qtmp,cfastLLy,nvar)
              qtmp(1)=qLR(1); qtmp(2)=qLR(2); qtmp(7)=qLR(5); qtmp(8)=qLR(8)
              qtmp(3)=qLR(4); qtmp(4)=qLR(7); qtmp(5)=qLR(3); qtmp(6)=qLR(6)
              call find_speed_fast(qtmp,cfastLRy,nvar)
              qtmp(1)=qRL(1); qtmp(2)=qRL(2); qtmp(7)=qRL(5); qtmp(8)=qRL(8)
              qtmp(3)=qRL(4); qtmp(4)=qRL(7); qtmp(5)=qRL(3); qtmp(6)=qRL(6)
              call find_speed_fast(qtmp,cfastRLy,nvar)
              qtmp(1)=qRR(1); qtmp(2)=qRR(2); qtmp(7)=qRR(5); qtmp(8)=qRR(8)
              qtmp(3)=qRR(4); qtmp(4)=qRR(7); qtmp(5)=qRR(3); qtmp(6)=qRR(6)
              call find_speed_fast(qtmp,cfastRRy,nvar)
              
              SL=min(uLL,uLR,uRL,uRR)-max(cfastLLx,cfastLRx,cfastRLx,cfastRRx)
              SR=max(uLL,uLR,uRL,uRR)+max(cfastLLx,cfastLRx,cfastRLx,cfastRRx)
              SB=min(vLL,vLR,vRL,vRR)-max(cfastLLy,cfastLRy,cfastRLy,cfastRRy)
              ST=max(vLL,vLR,vRL,vRR)+max(cfastLLy,cfastLRy,cfastRLy,cfastRRy)
              
              ELL=uLL*BLL-vLL*ALL
              ELR=uLR*BLR-vLR*ALR
              ERL=uRL*BRL-vRL*ARL
              ERR=uRR*BRR-vRR*ARR
              
              PtotLL=pLL+half*(ALL*ALL+BLL*BLL+CLL*CLL)
              PtotLR=pLR+half*(ALR*ALR+BLR*BLR+CLR*CLR)
              PtotRL=pRL+half*(ARL*ARL+BRL*BRL+CRL*CRL)
              PtotRR=pRR+half*(ARR*ARR+BRR*BRR+CRR*CRR)
              
              rcLLx=rLL*(uLL-SL); rcRLx=rRL*(SR-uRL) 
              rcLRx=rLR*(uLR-SL); rcRRx=rRR*(SR-uRR)
              rcLLy=rLL*(vLL-SB); rcLRy=rLR*(ST-vLR) 
              rcRLy=rRL*(vRL-SB); rcRRy=rRR*(ST-vRR)
              
              ustar=(rcLLx*uLL+rcLRx*uLR+rcRLx*uRL+rcRRx*uRR+(PtotLL-PtotRL+PtotLR-PtotRR))/(rcLLx+rcLRx+rcRLx+rcRRx)
              vstar=(rcLLy*vLL+rcLRy*vLR+rcRLy*vRL+rcRRy*vRR+(PtotLL-PtotLR+PtotRL-PtotRR))/(rcLLy+rcLRy+rcRLy+rcRRy)
              
              rstarLLx=rLL*(SL-uLL)/(SL-ustar); BstarLL=BLL*(SL-uLL)/(SL-ustar)
              rstarLLy=rLL*(SB-vLL)/(SB-vstar); AstarLL=ALL*(SB-vLL)/(SB-vstar)
              rstarLL =rLL*(SL-uLL)/(SL-ustar)*(SB-vLL)/(SB-vstar)
              EstarLLx=ustar*BstarLL-vLL  *ALL
              EstarLLy=uLL  *BLL    -vstar*AstarLL
              EstarLL =ustar*BstarLL-vstar*AstarLL
              
              rstarLRx=rLR*(SL-uLR)/(SL-ustar); BstarLR=BLR*(SL-uLR)/(SL-ustar)
              rstarLRy=rLR*(ST-vLR)/(ST-vstar); AstarLR=ALR*(ST-vLR)/(ST-vstar)
              rstarLR =rLR*(SL-uLR)/(SL-ustar)*(ST-vLR)/(ST-vstar)
              EstarLRx=ustar*BstarLR-vLR  *ALR
              EstarLRy=uLR  *BLR    -vstar*AstarLR
              EstarLR =ustar*BstarLR-vstar*AstarLR
              
              rstarRLx=rRL*(SR-uRL)/(SR-ustar); BstarRL=BRL*(SR-uRL)/(SR-ustar)
              rstarRLy=rRL*(SB-vRL)/(SB-vstar); AstarRL=ARL*(SB-vRL)/(SB-vstar)
              rstarRL =rRL*(SR-uRL)/(SR-ustar)*(SB-vRL)/(SB-vstar)
              EstarRLx=ustar*BstarRL-vRL  *ARL
              EstarRLy=uRL  *BRL    -vstar*AstarRL
              EstarRL =ustar*BstarRL-vstar*AstarRL
              
              rstarRRx=rRR*(SR-uRR)/(SR-ustar); BstarRR=BRR*(SR-uRR)/(SR-ustar)
              rstarRRy=rRR*(ST-vRR)/(ST-vstar); AstarRR=ARR*(ST-vRR)/(ST-vstar)
              rstarRR =rRR*(SR-uRR)/(SR-ustar)*(ST-vRR)/(ST-vstar)
              EstarRRx=ustar*BstarRR-vRR  *ARR
              EstarRRy=uRR  *BRR    -vstar*AstarRR
              EstarRR =ustar*BstarRR-vstar*AstarRR
              
              calfvenL=max(abs(ALR)/sqrt(rstarLRx),abs(AstarLR)/sqrt(rstarLR), &
                   &       abs(ALL)/sqrt(rstarLLx),abs(AstarLL)/sqrt(rstarLL),smallc)
              calfvenR=max(abs(ARR)/sqrt(rstarRRx),abs(AstarRR)/sqrt(rstarRR), &
                   &       abs(ARL)/sqrt(rstarRLx),abs(AstarRL)/sqrt(rstarRL),smallc)
              calfvenB=max(abs(BLL)/sqrt(rstarLLy),abs(BstarLL)/sqrt(rstarLL), &
                   &       abs(BRL)/sqrt(rstarRLy),abs(BstarRL)/sqrt(rstarRL),smallc)
              calfvenT=max(abs(BLR)/sqrt(rstarLRy),abs(BstarLR)/sqrt(rstarLR), &
                   &       abs(BRR)/sqrt(rstarRRy),abs(BstarRR)/sqrt(rstarRR),smallc)
              SAL=min(ustar-calfvenL,zero); SAR=max(ustar+calfvenR,zero)
              SAB=min(vstar-calfvenB,zero); SAT=max(vstar+calfvenT,zero)
              AstarT=(SAR*AstarRR-SAL*AstarLR)/(SAR-SAL); AstarB=(SAR*AstarRL-SAL*AstarLL)/(SAR-SAL)
              BstarR=(SAT*BstarRR-SAB*BstarRL)/(SAT-SAB); BstarL=(SAT*BstarLR-SAB*BstarLL)/(SAT-SAB)
              
              if(SB>0d0)then
                 if(SL>0d0)then
                    E=ELL
                 else if(SR<0d0)then
                    E=ERL
                 else
                    E=(SAR*EstarLLx-SAL*EstarRLx+SAR*SAL*(BRL-BLL))/(SAR-SAL)
                 endif
              else if (ST<0d0)then
                 if(SL>0d0)then
                    E=ELR
                 else if(SR<0d0)then
                    E=ERR
                 else
                    E=(SAR*EstarLRx-SAL*EstarRRx+SAR*SAL*(BRR-BLR))/(SAR-SAL)
                 endif
              else if(SL>0d0)then
                 E=(SAT*EstarLLy-SAB*EstarLRy-SAT*SAB*(ALR-ALL))/(SAT-SAB)
              else if(SR<0d0)then
                 E=(SAT*EstarRLy-SAB*EstarRRy-SAT*SAB*(ARR-ARL))/(SAT-SAB)
              else
                 E=(SAL*SAB*EstarRR-SAL*SAT*EstarRL-SAR*SAB*EstarLR+SAR*SAT*EstarLL)/(SAR-SAL)/(SAT-SAB) &
                      & -SAT*SAB/(SAT-SAB)*(AstarT-AstarB)+SAR*SAL/(SAR-SAL)*(BstarR-BstarL)
              endif
              
              emf(i,j,k) = E
              
           else if(riemann2d=='hllf')then
              
              ! Compute 4 fast magnetosonic velocity relative to x direction
              qtmp(1)=qLL(1); qtmp(2)=qLL(2); qtmp(7)=qLL(5); qtmp(8)=qLL(8)
              qtmp(3)=qLL(3); qtmp(4)=qLL(6); qtmp(5)=qLL(4); qtmp(6)=qLL(7)
              vLLx=qtmp(3); call find_speed_fast(qtmp,cLLx,nvar)
              qtmp(1)=qLR(1); qtmp(2)=qLR(2); qtmp(7)=qLR(5); qtmp(8)=qLR(8)
              qtmp(3)=qLR(3); qtmp(4)=qLR(6); qtmp(5)=qLR(4); qtmp(6)=qLR(7)
              vLRx=qtmp(3); call find_speed_fast(qtmp,cLRx,nvar)
              qtmp(1)=qRL(1); qtmp(2)=qRL(2); qtmp(7)=qRL(5); qtmp(8)=qRL(8)
              qtmp(3)=qRL(3); qtmp(4)=qRL(6); qtmp(5)=qRL(4); qtmp(6)=qRL(7)
              vRLx=qtmp(3); call find_speed_fast(qtmp,cRLx,nvar)
              qtmp(1)=qRR(1); qtmp(2)=qRR(2); qtmp(7)=qRR(5); qtmp(8)=qRR(8)
              qtmp(3)=qRR(3); qtmp(4)=qRR(6); qtmp(5)=qRR(4); qtmp(6)=qRR(7)
              vRRx=qtmp(3); call find_speed_fast(qtmp,cRRx,nvar)
              
              ! Compute 4 fast magnetosonic velocity relative to y direction
              qtmp(1)=qLL(1); qtmp(2)=qLL(2); qtmp(7)=qLL(5); qtmp(8)=qLL(8)
              qtmp(3)=qLL(4); qtmp(4)=qLL(7); qtmp(5)=qLL(3); qtmp(6)=qLL(6)
              vLLy=qtmp(3); call find_speed_fast(qtmp,cLLy,nvar)
              qtmp(1)=qLR(1); qtmp(2)=qLR(2); qtmp(7)=qLR(5); qtmp(8)=qLR(8)
              qtmp(3)=qLR(4); qtmp(4)=qLR(7); qtmp(5)=qLR(3); qtmp(6)=qLR(6)
              vLRy=qtmp(3); call find_speed_fast(qtmp,cLRy,nvar)
              qtmp(1)=qRL(1); qtmp(2)=qRL(2); qtmp(7)=qRL(5); qtmp(8)=qRL(8)
              qtmp(3)=qRL(4); qtmp(4)=qRL(7); qtmp(5)=qRL(3); qtmp(6)=qRL(6)
              vRLy=qtmp(3); call find_speed_fast(qtmp,cRLy,nvar)
              qtmp(1)=qRR(1); qtmp(2)=qRR(2); qtmp(7)=qRR(5); qtmp(8)=qRR(8)
              qtmp(3)=qRR(4); qtmp(4)=qRR(7); qtmp(5)=qRR(3); qtmp(6)=qRR(6)
              vRRy=qtmp(3); call find_speed_fast(qtmp,cRRy,nvar)
              
              SL=min(min(vLLx,vLRx,VRLx,vRRx)-max(cLLx,cLRx,cRLx,cRRx,smallc),zero)
              SR=max(max(vLLx,vLRx,VRLx,vRRx)+max(cLLx,cLRx,cRLx,cRRx,smallc),zero)
              SB=min(min(vLLy,vLRy,VRLy,vRRy)-max(cLLy,cLRy,cRLy,cRRy,smallc),zero)
              ST=max(max(vLLy,vLRy,VRLy,vRRy)+max(cLLy,cLRy,cRLy,cRRy,smallc),zero)
              
              emf(i,j,k) = (SL*SB*ERR-SL*ST*ERL-SR*SB*ELR+SR*ST*ELL)/(SR-SL)/(ST-SB) &
                   -ST*SB/(ST-SB)*(qRR(6)-qLL(6)) &
                   +SR*SL/(SR-SL)*(qRR(7)-qLL(7))
              
           else if (riemann2d=='hlla')then
              
              ! Compute 4 Alfven velocity relative to x direction
              qtmp(1)=qLL(1); qtmp(2)=qLL(2); qtmp(7)=qLL(5); qtmp(8)=qLL(8)
              qtmp(3)=qLL(3); qtmp(4)=qLL(6); qtmp(5)=qLL(4); qtmp(6)=qLL(7)
              vLLx=qtmp(3); call find_speed_alfven(qtmp,cLLx,nvar)
              qtmp(1)=qLR(1); qtmp(2)=qLR(2); qtmp(7)=qLR(5); qtmp(8)=qLR(8)
              qtmp(3)=qLR(3); qtmp(4)=qLR(6); qtmp(5)=qLR(4); qtmp(6)=qLR(7)
              vLRx=qtmp(3); call find_speed_alfven(qtmp,cLRx,nvar)
              qtmp(1)=qRL(1); qtmp(2)=qRL(2); qtmp(7)=qRL(5); qtmp(8)=qRL(8)
              qtmp(3)=qRL(3); qtmp(4)=qRL(6); qtmp(5)=qRL(4); qtmp(6)=qRL(7)
              vRLx=qtmp(3); call find_speed_alfven(qtmp,cRLx,nvar)
              qtmp(1)=qRR(1); qtmp(2)=qRR(2); qtmp(7)=qRR(5); qtmp(8)=qRR(8)
              qtmp(3)=qRR(3); qtmp(4)=qRR(6); qtmp(5)=qRR(4); qtmp(6)=qRR(7)
              vRRx=qtmp(3); call find_speed_alfven(qtmp,cRRx,nvar)
              ! Compute 4 Alfven relative to y direction
              qtmp(1)=qLL(1); qtmp(2)=qLL(2); qtmp(7)=qLL(5); qtmp(8)=qLL(8)
              qtmp(3)=qLL(4); qtmp(4)=qLL(7); qtmp(5)=qLL(3); qtmp(6)=qLL(6)
              vLLy=qtmp(3); call find_speed_alfven(qtmp,cLLy,nvar)
              qtmp(1)=qLR(1); qtmp(2)=qLR(2); qtmp(7)=qLR(5); qtmp(8)=qLR(8)
              qtmp(3)=qLR(4); qtmp(4)=qLR(7); qtmp(5)=qLR(3); qtmp(6)=qLR(6)
              vLRy=qtmp(3); call find_speed_alfven(qtmp,cLRy,nvar)
              qtmp(1)=qRL(1); qtmp(2)=qRL(2); qtmp(7)=qRL(5); qtmp(8)=qRL(8)
              qtmp(3)=qRL(4); qtmp(4)=qRL(7); qtmp(5)=qRL(3); qtmp(6)=qRL(6)
              vRLy=qtmp(3); call find_speed_alfven(qtmp,cRLy,nvar)
              qtmp(1)=qRR(1); qtmp(2)=qRR(2); qtmp(7)=qRR(5); qtmp(8)=qRR(8)
              qtmp(3)=qRR(4); qtmp(4)=qRR(7); qtmp(5)=qRR(3); qtmp(6)=qRR(6)
              vRRy=qtmp(3); call find_speed_alfven(qtmp,cRRy,nvar)
              
              SL=min(min(vLLx,vLRx,VRLx,vRRx)-max(cLLx,cLRx,cRLx,cRRx,smallc),zero)
              SR=max(max(vLLx,vLRx,VRLx,vRRx)+max(cLLx,cLRx,cRLx,cRRx,smallc),zero)
              SB=min(min(vLLy,vLRy,VRLy,vRRy)-max(cLLy,cLRy,cRLy,cRRy,smallc),zero)
              ST=max(max(vLLy,vLRy,VRLy,vRRy)+max(cLLy,cLRy,cRLy,cRRy,smallc),zero)
              
              emf(i,j,k) = (SL*SB*ERR-SL*ST*ERL-SR*SB*ELR+SR*ST*ELL)/(SR-SL)/(ST-SB) &
                   -ST*SB/(ST-SB)*(qRR(6)-qLL(6)) &
                   +SR*SL/(SR-SL)*(qRR(7)-qLL(7))
              
           else
              
              ! find the average value of E
              E = forth*(ELL+ERL+ELR+ERR)
              
              ! call the first solver in the x direction
              ! density
              qleft (1) = half*(qLL(1)+qLR(1))
              qright(1) = half*(qRR(1)+qRL(1))
              
              ! pressure
              qleft (2) = half*(qLL(2)+qLR(2))
              qright(2) = half*(qRR(2)+qRL(2))
              
              ! vt1 becomes normal velocity
              qleft (3) = half*(qLL(3)+qLR(3))
              qright(3) = half*(qRR(3)+qRL(3))
              
              ! bt1 becomes normal magnetic field
              qleft (4) = half*(qLL(6)+qLR(6))
              qright(4) = half*(qRR(6)+qRL(6))
              
              ! vt2 becomes transverse velocity field
              qleft (5) = half*(qLL(4)+qLR(4))
              qright(5) = half*(qRR(4)+qRL(4))
              
              ! bt2 becomes transverse magnetic field 
              qleft (6) = half*(qLL(7)+qLR(7))
              qright(6) = half*(qRR(7)+qRL(7))
              
              ! velocity component perp. to the plane is now transverse
              qleft (7) = half*(qLL(5)+qLR(5))
              qright(7) = half*(qRR(5)+qRL(5))
              
              ! magnetic field component perp. to the plane is now transverse
              qleft (8) = half*(qLL(8)+qLR(8))
              qright(8) = half*(qRR(8)+qRL(8))
              
              zero_flux = 0.0
              select case (riemann2d)
              case ('roe')
                 call athena_roe   (qleft,qright,fmean_x,zero_flux)
              case ('llf')
                 call lax_friedrich(qleft,qright,fmean_x,zero_flux)
              case ('upwind')
                 call upwind       (qleft,qright,fmean_x,zero_flux)
              end select
              
              ! call the second solver in the y direction
              ! density
              qleft (1) = half*(qLL(1)+qRL(1))
              qright(1) = half*(qRR(1)+qLR(1))
              
              ! pressure
              qleft (2) = half*(qLL(2)+qRL(2))
              qright(2) = half*(qRR(2)+qLR(2))
              
              ! vt2 becomes normal velocity
              qleft (3) = half*(qLL(4)+qRL(4))
              qright(3) = half*(qRR(4)+qLR(4))
              
              ! bt2 becomes normal magnetic field
              qleft (4) = half*(qLL(7)+qRL(7))
              qright(4) = half*(qRR(7)+qLR(7))
              
              ! vt1 becomes transverse velocity field 
              qleft (5) = half*(qLL(3)+qRL(3))
              qright(5) = half*(qRR(3)+qLR(3))
              
              ! bt1 becomes transverse magnetic field 
              qleft (6) = half*(qLL(6)+qRL(6))
              qright(6) = half*(qRR(6)+qLR(6))
              
              ! velocity component perp. to the plane is now transverse
              qleft (7) = half*(qLL(5)+qRL(5))
              qright(7) = half*(qRR(5)+qLR(5))
              
              ! magnetic field component perp. to the plane is now transverse
              qleft (8) = half*(qLL(8)+qRL(8))
              qright(8) = half*(qRR(8)+qLR(8))
              
              zero_flux = 0.
              select case (riemann2d)
              case ('roe')
                 call athena_roe   (qleft,qright,fmean_y,zero_flux)
              case ('llf')
                 call lax_friedrich(qleft,qright,fmean_y,zero_flux)
              case ('upwind')
                 call upwind       (qleft,qright,fmean_y,zero_flux)
              end select
              
              ! compute the final value of E including the 2D diffusive
              ! terms that ensure stability 
              emf(i,j,k) = E + (fmean_x(6) - fmean_y(6))
              
           endif
           
        end do
     end do
  end do

  deallocate(qLL,qRL,qLR,qRR)
  deallocate(qleft,qright,fmean_x,fmean_y,qtmp)
  
  return

end subroutine cmp_mag_flx

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine FAST_MHD_SPEED
!
!> Computes the fast magnetosonic velocity.
!<
subroutine fast_mhd_speed(rho,p,vx,vy,vz,bx,by,bz,fast_speed)

  use gammas
  use const

  implicit none

  real :: rho,p,vx,vy,vz,bx,by,bz
  real :: mag_perp,alfv,vit_son,som_vit,delta,fast_speed

  mag_perp =  (by**2+bz**2) /  rho  !bt ^2 / rho
  alfv     =  bx**2         /  rho  ! bx / sqrt(4pi*rho)
  vit_son  =  gamma*p       /  rho  !sonic contribution :  gamma*P / rho

  som_vit  =  mag_perp + alfv + vit_son 
  
  delta   = max(zero, som_vit**2 - four * vit_son*alfv ) 
  
  fast_speed =  half*( som_vit + sqrt( delta )   )
  fast_speed =  sqrt(fast_speed)

  return

end subroutine fast_mhd_speed

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine USLOPE
!
!> Computes mhd slopes.
!<
subroutine uslope(bf,q,dq,dbf,iu1,iu2,ju1,ju2,ku1,ku2,nvar,ndim)

  use parameters, only : slope_type
  use const
  use para      , only : mype

  implicit none

  integer :: nvar,ndim
  integer :: iu1,iu2,ju1,ju2,ku1,ku2

  real, dimension(iu1:iu2+1,ju1:ju2+1,ku1:ku2+1,1:3          ) :: bf
  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:nvar       ) :: q 
  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:nvar,1:ndim) :: dq
  real, dimension(iu1:iu2  ,ju1:ju2  ,ku1:ku2  ,1:3   ,1:ndim) :: dbf

  ! local variables
  integer :: i,j,k,n
  integer :: ilo,ihi,jlo,jhi,klo,khi
  real    :: dsgn,dlim,dcen,dlft,drgt,slop
  real    :: dfll,dflm,dflr,dfml,dfmm,dfmr,dfrl,dfrm,dfrr
  real    :: dflll,dflml,dflrl,dfmll,dfmml,dfmrl,dfrll,dfrml,dfrrl
  real    :: dfllm,dflmm,dflrm,dfmlm,dfmmm,dfmrm,dfrlm,dfrmm,dfrrm
  real    :: dfllr,dflmr,dflrr,dfmlr,dfmmr,dfmrr,dfrlr,dfrmr,dfrrr
  real    :: vmin,vmax,dfx,dfy,dfz,dff,xslope_type
  
  ilo=min(1,iu1+1); ihi=max(1,iu2-1)
  jlo=min(1,ju1+1); jhi=max(1,ju2-1)
  klo=min(1,ku1+1); khi=max(1,ku2-1)

  if(slope_type==0)then
     dq=zero
     dbf=zero
     return
  end if

  !1D
  if (ndim==1) then
     do n = 1, nvar
        do k = klo, khi
           do j = jlo, jhi
              do i = ilo, ihi
                 if(slope_type==1.or.slope_type==2)then  ! minmod or average
                    dlft = slope_type*(q(i  ,j,k,n) - q(i-1,j,k,n))
                    drgt = slope_type*(q(i+1,j,k,n) - q(i  ,j,k,n))
                    dcen = half*(dlft+drgt)/slope_type
                    dsgn = sign(one, dcen)
                    slop = min(abs(dlft),abs(drgt))
                    dlim = slop
                    if((dlft*drgt)<=zero)dlim=zero
                    dq(i,j,k,n,1) = dsgn*min(dlim,abs(dcen))
                 else
                    call close_program(mype,'Unknown slope type')
                 end if
              end do
           end do
        end do
     end do
  endif

  !2D
  if (ndim==2) then
     if(slope_type==1.or.slope_type==2)then  ! minmod or average
        do n = 1, nvar
           do k = klo, khi
              do j = jlo, jhi
                 do i = ilo, ihi
                    ! slopes in first coordinate direction
                    dlft = slope_type*(q(i  ,j,k,n) - q(i-1,j,k,n))
                    drgt = slope_type*(q(i+1,j,k,n) - q(i  ,j,k,n))
                    dcen = half*(dlft+drgt)/slope_type
                    dsgn = sign(one, dcen)
                    slop = min(abs(dlft),abs(drgt))
                    dlim = slop
                    if((dlft*drgt)<=zero)dlim=zero
                    dq(i,j,k,n,1) = dsgn*min(dlim,abs(dcen))
                    ! slopes in second coordinate direction
                    dlft = slope_type*(q(i,j  ,k,n) - q(i,j-1,k,n))
                    drgt = slope_type*(q(i,j+1,k,n) - q(i,j  ,k,n))
                    dcen = half*(dlft+drgt)/slope_type
                    dsgn = sign(one,dcen)
                    slop = min(abs(dlft),abs(drgt))
                    dlim = slop
                    if((dlft*drgt)<=zero)dlim=zero
                    dq(i,j,k,n,2) = dsgn*min(dlim,abs(dcen))
                 end do
              end do
           end do
        end do
        
     else if(slope_type==3)then ! positivity preserving 2d unsplit slope
        
        do n = 1, nvar
           do k = klo, khi
              do j = jlo, jhi
                 do i = ilo, ihi
                    dfll = q(i-1,j-1,k,n)-q(i,j,k,n)
                    dflm = q(i-1,j  ,k,n)-q(i,j,k,n)
                    dflr = q(i-1,j+1,k,n)-q(i,j,k,n)
                    dfml = q(i  ,j-1,k,n)-q(i,j,k,n)
                    dfmm = q(i  ,j  ,k,n)-q(i,j,k,n)
                    dfmr = q(i  ,j+1,k,n)-q(i,j,k,n)
                    dfrl = q(i+1,j-1,k,n)-q(i,j,k,n)
                    dfrm = q(i+1,j  ,k,n)-q(i,j,k,n)
                    dfrr = q(i+1,j+1,k,n)-q(i,j,k,n)
                    
                    vmin = min(dfll,dflm,dflr,dfml,dfmm,dfmr,dfrl,dfrm,dfrr)
                    vmax = max(dfll,dflm,dflr,dfml,dfmm,dfmr,dfrl,dfrm,dfrr)
                    
                    dfx  = half*(q(i+1,j,k,n)-q(i-1,j,k,n))
                    dfy  = half*(q(i,j+1,k,n)-q(i,j-1,k,n))
                    dff  = half*(abs(dfx)+abs(dfy))
                    
                    if(dff>zero)then
                       slop = min(one,min(abs(vmin),abs(vmax))/dff)
                    else
                       slop = one
                    endif
                    
                    dlim = slop
                    
                    dq(i,j,k,n,1) = dlim*dfx
                    dq(i,j,k,n,2) = dlim*dfy
                    
                 end do
              end do
           end do
        end do
     else
        call close_program(mype,'Unknown slope type')
     endif
     
     ! 1D transverse TVD slopes for face-centered magnetic fields
     ! Bx along direction Y 
     do k = klo, khi
        do j = jlo, jhi
           do i = ilo, ihi+1 ! WARNING HERE
              dlft = slope_type*(bf(i,j  ,k,1) - bf(i,j-1,k,1))
              drgt = slope_type*(bf(i,j+1,k,1) - bf(i,j  ,k,1))
              dcen = half*(dlft+drgt)/slope_type
              dsgn = sign(one, dcen)
              slop = min(abs(dlft),abs(drgt))
              dlim = slop
              if((dlft*drgt)<=zero)dlim=zero
              dbf(i,j,k,1,2) = dsgn*min(dlim,abs(dcen))
           enddo
        end do
     end do
     ! By along direction X
     do k = klo, khi
        do j = jlo, jhi+1 ! WARNING HERE
           do i = ilo, ihi
              dlft = slope_type*(bf(i  ,j,k,2) - bf(i-1,j,k,2))
              drgt = slope_type*(bf(i+1,j,k,2) - bf(i  ,j,k,2))
              dcen = half*(dlft+drgt)/slope_type
              dsgn = sign(one, dcen)
              slop = min(abs(dlft),abs(drgt))
              dlim = slop
              if((dlft*drgt)<=zero)dlim=zero
              dbf(i,j,k,2,1) = dsgn*min(dlim,abs(dcen))
           enddo
        end do
     end do
  endif

  !3D
  if (ndim==3) then
     if(slope_type==1.or.slope_type==2)then  ! minmod or average
        do n = 1, nvar
           do k = klo, khi
              do j = jlo, jhi
                 do i = ilo, ihi
                    ! slopes in first coordinate direction
                    dlft = slope_type*(q(i  ,j,k,n) - q(i-1,j,k,n))
                    drgt = slope_type*(q(i+1,j,k,n) - q(i  ,j,k,n))
                    dcen = half*(dlft+drgt)/slope_type
                    dsgn = sign(one, dcen)
                    slop = min(abs(dlft),abs(drgt))
                    dlim = slop
                    if((dlft*drgt)<=zero)dlim=zero
                    dq(i,j,k,n,1) = dsgn*min(dlim,abs(dcen))
                    ! slopes in second coordinate direction
                    dlft = slope_type*(q(i,j  ,k,n) - q(i,j-1,k,n))
                    drgt = slope_type*(q(i,j+1,k,n) - q(i,j  ,k,n))
                    dcen = half*(dlft+drgt)/slope_type
                    dsgn = sign(one,dcen)
                    slop = min(abs(dlft),abs(drgt))
                    dlim = slop
                    if((dlft*drgt)<=zero)dlim=zero
                    dq(i,j,k,n,2) = dsgn*min(dlim,abs(dcen))
                    ! slopes in third coordinate direction
                    dlft = slope_type*(q(i,j,k  ,n) - q(i,j,k-1,n))
                    drgt = slope_type*(q(i,j,k+1,n) - q(i,j,k  ,n))
                    dcen = half*(dlft+drgt)/slope_type
                    dsgn = sign(one,dcen)
                    slop = min(abs(dlft),abs(drgt))
                    dlim = slop
                    if((dlft*drgt)<=zero)dlim=zero
                    dq(i,j,k,n,3) = dsgn*min(dlim,abs(dcen))
                 end do
              end do
           end do
        end do
        
     else if(slope_type==3)then ! positivity preserving 3d unsplit slope
        do n = 1, nvar
           do k = klo, khi
              do j = jlo, jhi
                 do i = ilo, ihi
                    dflll = q(i-1,j-1,k-1,n)-q(i,j,k,n)
                    dflml = q(i-1,j  ,k-1,n)-q(i,j,k,n)
                    dflrl = q(i-1,j+1,k-1,n)-q(i,j,k,n)
                    dfmll = q(i  ,j-1,k-1,n)-q(i,j,k,n)
                    dfmml = q(i  ,j  ,k-1,n)-q(i,j,k,n)
                    dfmrl = q(i  ,j+1,k-1,n)-q(i,j,k,n)
                    dfrll = q(i+1,j-1,k-1,n)-q(i,j,k,n)
                    dfrml = q(i+1,j  ,k-1,n)-q(i,j,k,n)
                    dfrrl = q(i+1,j+1,k-1,n)-q(i,j,k,n)
                    
                    dfllm = q(i-1,j-1,k  ,n)-q(i,j,k,n)
                    dflmm = q(i-1,j  ,k  ,n)-q(i,j,k,n)
                    dflrm = q(i-1,j+1,k  ,n)-q(i,j,k,n)
                    dfmlm = q(i  ,j-1,k  ,n)-q(i,j,k,n)
                    dfmmm = q(i  ,j  ,k  ,n)-q(i,j,k,n)
                    dfmrm = q(i  ,j+1,k  ,n)-q(i,j,k,n)
                    dfrlm = q(i+1,j-1,k  ,n)-q(i,j,k,n)
                    dfrmm = q(i+1,j  ,k  ,n)-q(i,j,k,n)
                    dfrrm = q(i+1,j+1,k  ,n)-q(i,j,k,n)
                    
                    dfllr = q(i-1,j-1,k+1,n)-q(i,j,k,n)
                    dflmr = q(i-1,j  ,k+1,n)-q(i,j,k,n)
                    dflrr = q(i-1,j+1,k+1,n)-q(i,j,k,n)
                    dfmlr = q(i  ,j-1,k+1,n)-q(i,j,k,n)
                    dfmmr = q(i  ,j  ,k+1,n)-q(i,j,k,n)
                    dfmrr = q(i  ,j+1,k+1,n)-q(i,j,k,n)
                    dfrlr = q(i+1,j-1,k+1,n)-q(i,j,k,n)
                    dfrmr = q(i+1,j  ,k+1,n)-q(i,j,k,n)
                    dfrrr = q(i+1,j+1,k+1,n)-q(i,j,k,n)
                    
                    vmin = min(dflll,dflml,dflrl,dfmll,dfmml,dfmrl,dfrll,dfrml,dfrrl, &
                         &     dfllm,dflmm,dflrm,dfmlm,dfmmm,dfmrm,dfrlm,dfrmm,dfrrm, &
                         &     dfllr,dflmr,dflrr,dfmlr,dfmmr,dfmrr,dfrlr,dfrmr,dfrrr)
                    vmax = max(dflll,dflml,dflrl,dfmll,dfmml,dfmrl,dfrll,dfrml,dfrrl, &
                         &     dfllm,dflmm,dflrm,dfmlm,dfmmm,dfmrm,dfrlm,dfrmm,dfrrm, &
                         &     dfllr,dflmr,dflrr,dfmlr,dfmmr,dfmrr,dfrlr,dfrmr,dfrrr)
                    
                    dfx  = half*(q(i+1,j,k,n)-q(i-1,j,k,n))
                    dfy  = half*(q(i,j+1,k,n)-q(i,j-1,k,n))
                    dfz  = half*(q(i,j,k+1,n)-q(i,j,k-1,n))
                    dff  = half*(abs(dfx)+abs(dfy)+abs(dfz))
                    
                    if(dff>zero)then
                       slop = min(one,min(abs(vmin),abs(vmax))/dff)
                    else
                       slop = one
                    endif
                    
                    dlim = slop
                    
                    dq(i,j,k,n,1) = dlim*dfx
                    dq(i,j,k,n,2) = dlim*dfy
                    dq(i,j,k,n,3) = dlim*dfz
                    
                 end do
              end do
           end do
        end do
     else
        call close_program(mype,'Unknown slope type')
     endif
     
     ! 2D transverse TVD slopes for face-centered magnetic fields
     ! Bx along direction Y and Z
     xslope_type=min(slope_type,2)
     if(slope_type==1.or.slope_type==2.or.slope_type==3)then  ! minmod or average
        do k = klo, khi
           do j = jlo, jhi
              do i = ilo, ihi+1 ! WARNING HERE
                 ! slopes along Y
                 dlft = xslope_type*(bf(i,j  ,k,1) - bf(i,j-1,k,1))
                 drgt = xslope_type*(bf(i,j+1,k,1) - bf(i,j  ,k,1))
                 dcen = half*(dlft+drgt)/xslope_type
                 dsgn = sign(one, dcen)
                 slop = min(abs(dlft),abs(drgt))
                 dlim = slop
                 if((dlft*drgt)<=zero)dlim=zero
                 dbf(i,j,k,1,2) = dsgn*min(dlim,abs(dcen))
                 ! slopes along Z
                 dlft = xslope_type*(bf(i,j,k  ,1) - bf(i,j,k-1,1))
                 drgt = xslope_type*(bf(i,j,k+1,1) - bf(i,j,k  ,1))
                 dcen = half*(dlft+drgt)/xslope_type
                 dsgn = sign(one,dcen)
                 slop = min(abs(dlft),abs(drgt))
                 dlim = slop
                 if((dlft*drgt)<=zero)dlim=zero
                 dbf(i,j,k,1,3) = dsgn*min(dlim,abs(dcen))
              end do
           end do
        end do
     endif
     ! By along direction X and Z
     if(slope_type==1.or.slope_type==2.or.slope_type==3)then  ! minmod or average
        do k = klo, khi
           do j = jlo, jhi+1 ! WARNING HERE
              do i = ilo, ihi
                 ! slopes along X
                 dlft = xslope_type*(bf(i  ,j,k,2) - bf(i-1,j,k,2))
                 drgt = xslope_type*(bf(i+1,j,k,2) - bf(i  ,j,k,2))
                 dcen = half*(dlft+drgt)/xslope_type
                 dsgn = sign(one, dcen)
                 slop = min(abs(dlft),abs(drgt))
                 dlim = slop
                 if((dlft*drgt)<=zero)dlim=zero
                 dbf(i,j,k,2,1) = dsgn*min(dlim,abs(dcen))
                 ! slopes along Z
                 dlft = xslope_type*(bf(i,j,k  ,2) - bf(i,j,k-1,2))
                 drgt = xslope_type*(bf(i,j,k+1,2) - bf(i,j,k  ,2))
                 dcen = half*(dlft+drgt)/xslope_type
                 dsgn = sign(one,dcen)
                 slop = min(abs(dlft),abs(drgt))
                 dlim = slop
                 if((dlft*drgt)<=zero)dlim=zero
                 dbf(i,j,k,2,3) = dsgn*min(dlim,abs(dcen))
              end do
           end do
        end do
     endif
     ! Bz along direction X and Y
     if(slope_type==1.or.slope_type==2.or.slope_type==3)then  ! minmod or average
        do k = klo, khi+1 ! WARNING HERE
           do j = jlo, jhi
              do i = ilo, ihi
                 ! slopes along X
                 dlft = xslope_type*(bf(i  ,j,k,3) - bf(i-1,j,k,3))
                 drgt = xslope_type*(bf(i+1,j,k,3) - bf(i  ,j,k,3))
                 dcen = half*(dlft+drgt)/xslope_type
                 dsgn = sign(one, dcen)
                 slop = min(abs(dlft),abs(drgt))
                 dlim = slop
                 if((dlft*drgt)<=zero)dlim=zero
                 dbf(i,j,k,3,1) = dsgn*min(dlim,abs(dcen))
                 ! slopes along Y
                 dlft = xslope_type*(bf(i,j  ,k,3) - bf(i,j-1,k,3))
                 drgt = xslope_type*(bf(i,j+1,k,3) - bf(i,j  ,k,3))
                 dcen = half*(dlft+drgt)/xslope_type
                 dsgn = sign(one,dcen)
                 slop = min(abs(dlft),abs(drgt))
                 dlim = slop
                 if((dlft*drgt)<=zero)dlim=zero
                 dbf(i,j,k,3,2) = dsgn*min(dlim,abs(dcen))
              end do
           end do
        end do
     endif
  endif
  
  return
  
end subroutine uslope
