!> \file
!! Contains subroutines diffusion_imp(), mat_prod_dif()
!! and cmp_precond_dif()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine DIFFUSION_IMP
!
!> This routine solves the FLD equations implicitely with 
!! a conjugate gradient method.
!<
subroutine diffusion_imp
  
  use var
  use geom
  use divers    , only : verbose,dt
  use parameters
  use cl_dif
  use para      , only : mype
  use diffusion
  use unites    , only : c
  
  implicit none

#if PARA==1
  include 'mpif.h'
  integer :: info
#endif

  integer               :: i,j,k,nite
  
  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: TT,BB
  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: Temp,cv,kappa_abs,kappa_sca
  real                        :: dif,varmax,var_loc,xx,eps
  real                        :: omega,den,source_dif,derive_source_dif
  
  interface
     subroutine mat_prod_dif(X,AX)
       use parameters, only : nxmin, nxmax
       implicit none
       real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: X, AX
     end subroutine mat_prod_dif
     subroutine cmp_precond_dif(R,Z)
       use parameters, only : nxmin, nxmax
       implicit none
       real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: R, Z
     end subroutine cmp_precond_dif
  end interface

  if (verbose) write(*,*) 'Start of diffusion_imp'

  ! compute T
  if(verbose) write(*,*) 'compute  T'

  E = E - 0.5*sum(rhou**2,4)/rho
  if (allocated(B)) E = E - 0.5*sum(B**2,4)
  call temperature  (rho,E   ,Temp,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))
  call heat_capacity(rho,Temp,cv  ,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))
  E = E + 0.5*sum(rhou**2,4)/rho
  if (allocated(B)) E = E + 0.5*sum(B**2,4)

  ! Initialize unknown vector
  TT = Er_dif

  ! Initialize right hand side
  call cal_kappa_dif(rho,Temp,kappa_abs,kappa_sca,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))

  do k = nxmin(3),nxmax(3)
     do j = nxmin(2),nxmax(2)
        do i = nxmin(1),nxmax(1)
           omega = kappa_abs(i,j,k)*rho(i,j,k)*c*dt
           den   = cv(i,j,k)+omega*derive_source_dif(Temp(i,j,k))

           BB(i,j,k) = Er_dif(i,j,k)+cv(i,j,k)*omega*source_dif(Temp(i,j,k))/den
        enddo
     enddo
  enddo

  if(minval(Temp(:,:,:))<0.) then
     do k = nxmin(3),nxmax(3)
     do j = nxmin(2),nxmax(2)
     do i = nxmin(1),nxmax(1)
        if(Temp(i,j,k)<0.) then
           write(*,101) i,j,k,Temp(i,j,k)
101        format('Tloc negative (i,j,k), T=',I5,I5,I5,2x,E10.3)
           call temperature(rho,1.e-8*E,xx,1,1,1,1,1,1)
           Temp(i,j,k) = xx
        endif
     enddo
     enddo
     enddo
  endif

  ! Conjugate Gradient method
  eps = 1.e-6
  call cg(TT,BB,icl_dif,mat_prod_dif,precond_dif,cmp_precond_dif,eps,nite,dif)
  nitetot_dif=nitetot_dif+nite

  ! Update
  
  var_loc = maxval(2.*abs(TT-Er_dif)/(TT+Er_dif))

  Er_dif = TT

  ! Update T
  do k = nxmin(3),nxmax(3)
     do j = nxmin(2),nxmax(2)
        do i = nxmin(1),nxmax(1)
           omega = kappa_abs(i,j,k)*rho(i,j,k)*c*dt
           den   = cv(i,j,k)+omega*derive_source_dif(Temp(i,j,k))

           Temp(i,j,k) = (cv(i,j,k)*Temp(i,j,k)+omega*(Er_dif(i,j,k)- &
                         source_dif(Temp(i,j,k))+derive_source_dif(Temp(i,j,k))*Temp(i,j,k))) / den
        enddo
     enddo
  enddo

  call energy(rho,Temp,E,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))
  E = E + 0.5*sum(rhou**2,4)/rho
  if (allocated(B)) E = E + 0.5*sum(B**2,4)

#if PARA==1
  info = 0 
  call MPI_ALLREDUCE(var_loc, varmax ,1,MPI_DOUBLE_PRECISION,MPI_MAX,MPI_COMM_WORLD,info)
#else
  varmax = var_loc
#endif
  
  if (mype == 0) write(*,997) nite,varmax*100.,dif,nitetot_dif
997 format('nite =',I4,' varmax =',1pe9.2,'%   err =',E10.2,'    nitetot =',I8)

  dt_dif_imp  = vardt_dif*varrel_dif*dt/Max(vardt_dif*varmax,varrel_dif)

  if (verbose) write(*,*) 'End of diffusion_imp'

  return
  
end subroutine diffusion_imp

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine MAT_PROD_DIF
!
!> This routine computes the matric product A * X.
!<
subroutine mat_prod_dif(T,AT)
  
  use var
  use var_loc   , only : Tloc,xloc
  use geom
  use divers
  use parameters
  use unites    , only : c
  use diffusion , only : fld_limiter,er_dif

  implicit none
  
  integer, dimension(3,3) :: kron
  integer                 :: nnx,nny,nnz,nn
  integer                 :: i,j,k,idim,id
  integer                 :: ip1,jp1,kp1,im1,jm1,km1

  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: T, AT
  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: kappa_abs,kappa_sca,cv
  real, dimension(:,:,:,:  ), allocatable                                :: k_gradE
  real :: sig,flux_ener_tot,xx,lambda_fld,coeff,omega,derive_source_dif

  if (verbose) write(*,*) 'Begin of Mat_prod_dif'

  ! Arrays allocation
  nnx = nx(1) ; nny = nx(2) ; nnz = nx(3)

  kron = 0
  do idim = 1,3
     kron(idim,idim) = 1
  enddo

  nn  = max(nnx,nny,nnz)
  allocate(xloc(-1:nn+3,ndim)) 

  allocate(Tloc   (nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)       ))
  allocate(k_gradE(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3),1:ndim))

                 xloc(-1:nnx+3,1) = x(-1:nnx+3,1)   
  if (ndim.gt.1) xloc(-1:nny+3,2) = x(-1:nny+3,2)   
  if (ndim.gt.2) xloc(-1:nnz+3,3) = x(-1:nnz+3,3)   

  ! compute the geometry
  if(verbose) write(*,*) 'compute the geometry'
  call cal_geom(nnx,nny,nnz)

  ! compute kappa
  if(verbose) write(*,*) 'compute kappa'

  E = E - 0.5*sum(rhou**2,4)/rho
  if (allocated(B)) E = E - 0.5*sum(B**2,4)
  call   temperature(rho,E   ,Tloc,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))
  call heat_capacity(rho,Tloc,cv  ,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))
  E = E + 0.5*sum(rhou**2,4)/rho
  if (allocated(B)) E = E + 0.5*sum(B**2,4)

  call cal_kappa_dif(rho,Tloc,kappa_abs,kappa_sca,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))

  ! compute k_gradE
  do idim = 1,ndim
  
     do k = 1,nnz+kron(idim,3)
     do j = 1,nny+kron(idim,2)
     do i = 1,nnx+kron(idim,1)

        im1 = i - kron(1,idim) ; jm1 = j - kron(2,idim) ; km1 = k - kron(3,idim)

        coeff = 1.
        if(idim.eq.1) then
           id = i
        else if(idim.eq.2) then
           id = j
           if (geom_dir(2) == 'phi_c' .or. geom_dir(2) == 'the_s') coeff = 0.5*(x(i-1,1)+x(i,1))
        else if(idim.eq.3) then
           id = k
           if (geom_dir(3) == 'phi_s') coeff = 0.5*(x(i-1,1)+x(i,1))*sin(0.5*(x(j-1,2)+x(j,2)))
        endif

        sig = 0.5*(kappa_abs(im1,jm1,km1)*rho(im1,jm1,km1) + kappa_abs(i,j,k)*rho(i,j,k))

        !FLD 
        ! must use xx with the Er_dif(i,j,k) variable and not T(i,j,k)
        !(if not, some spurious oscillations appear even for equilibrium E=aT4)
        xx = max(1.e-7,abs( (Er_dif(i,j,k) - Er_dif(im1,jm1,km1)) / &
             (0.5*(Er_dif(i,j,k) + Er_dif(im1,jm1,km1))) ) )
        xx = xx/(0.5*(x(id+1,idim)-x(id-1,idim))*coeff) / sig

        k_gradE(i,j,k,idim) = c/sig * (T(i,j,k) - T(im1,jm1,km1)) / &
                              (0.5*(xloc(id+1,idim)-xloc(id-1,idim))*coeff) * lambda_fld(xx)

     enddo
     enddo
     enddo

  enddo


  AT = T

  do idim  = 1,ndim
  
     do k = 1,nnz
     do j = 1,nny
     do i = 1,nnx
     
        ip1 = i + kron(1,idim) ; jp1 = j + kron(2,idim) ; kp1 = k + kron(3,idim)

        flux_ener_tot  = k_gradE(ip1,jp1,kp1,idim)*ds(ip1,jp1,kp1,idim) -  k_gradE(i,j,k,idim)*ds(i,j,k,idim)

        AT(i,j,k) = AT(i,j,k) - flux_ener_tot * dt / dv(i,j,k)
        
     enddo
     enddo
     enddo
     
  enddo

  do k = 1,nnz
     do j = 1,nny
        do i = 1,nnx
           omega = kappa_abs(i,j,k)*rho(i,j,k)*c*dt
           AT(i,j,k) = AT(i,j,k) + T(i,j,k) * omega*cv(i,j,k)/(cv(i,j,k)+omega*derive_source_dif(Tloc(i,j,k)))
        enddo
     enddo
  enddo

  ! Arrays deallocation
  deallocate(dx,dxc)
  deallocate(ds,dv)
  deallocate(xloc,xcloc)
  deallocate(Tloc)
  deallocate(k_gradE)

  if (verbose) write(*,*) 'End of Mat_prod_dif'
  
  return

end subroutine mat_prod_dif

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine CMP_PRECOND_DIF
!
!> This routine computes the preconditioned matrix.
!<
subroutine cmp_precond_dif(RR,ZZ)
  
  use var
  use var_loc   , only : Tloc,xloc
  use geom
  use divers
  use parameters
  use unites    , only : c
  use diffusion , only : fld_limiter,er_dif

  implicit none
  
  integer, dimension(3,3) :: kron
  integer                 :: nnx,nny,nnz,nn
  integer                 :: i,j,k,idim,id,iface
  integer                 :: ip,jp,kp,im,jm,km

  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: RR,ZZ
  real, dimension(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)) :: kappa_abs,kappa_sca,cv
  real :: sig,xx,lambda_fld,coeff,omega,derive_source_dif,norm

  if (verbose) write(*,*) 'Start of cmp_precond_dif'

  ! Arrays allocation
  nnx = nx(1) ; nny = nx(2) ; nnz = nx(3)

  kron = 0
  do idim = 1,3
     kron(idim,idim) = 1
  enddo

  nn = max(nnx,nny,nnz)
  allocate(xloc(-1:nn+3,ndim)) 

  allocate(Tloc(nxmin(1):nxmax(1),nxmin(2):nxmax(2),nxmin(3):nxmax(3)))

                 xloc(-1:nnx+3,1) = x(-1:nnx+3,1)   
  if (ndim.gt.1) xloc(-1:nny+3,2) = x(-1:nny+3,2)   
  if (ndim.gt.2) xloc(-1:nnz+3,3) = x(-1:nnz+3,3)   

  ! compute the geometry
  if(verbose) write(*,*) 'compute the geometry'
  call cal_geom(nnx,nny,nnz)

  ! compute kappa
  if(verbose) write(*,*) 'compute kappa'

  E = E - 0.5*sum(rhou**2,4)/rho
  if (allocated(B)) E = E - 0.5*sum(B**2,4)
  call temperature  (rho,E   ,Tloc,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))
  call heat_capacity(rho,Tloc,cv  ,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))
  E = E + 0.5*sum(rhou**2,4)/rho
  if (allocated(B)) E = E + 0.5*sum(B**2,4)

  call cal_kappa_dif(rho,Tloc,kappa_abs,kappa_sca,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))

  ! compute the diagonal element of the matrix and normalize by it
  do k = 1,nnz
     do j = 1,nny
        do i = 1,nnx
           omega = kappa_abs(i,j,k)*rho(i,j,k)*c*dt
           
           ! term of dtE and source term
           norm = 1.+omega*cv(i,j,k)/(cv(i,j,k)+omega*derive_source_dif(Tloc(i,j,k)))

           ! term of div.F
           do iface = 1,2
           do idim = 1,ndim
           
              if (iface.eq.1) then
                 ip = i                ; jp = j                ; kp = k
                 im = i - kron(1,idim) ; jm = j - kron(2,idim) ; km = k - kron(3,idim)
              else
                 ip = i + kron(1,idim) ; jp = j + kron(2,idim) ; kp = k + kron(3,idim)
                 im = i                ; jm = j                ; km = k
              endif

              coeff=1.
              if(idim.eq.1) then
                 id = ip
              else if(idim.eq.2) then
                 id = jp
                 if (geom_dir(2) == 'phi_c' .or. geom_dir(2) == 'the_s') coeff = 0.5*(x(im,1)+x(ip,1))
              else if(idim.eq.3) then
                 id = kp
                 if (geom_dir(3) == 'phi_s') coeff = 0.5*(x(im,1)+x(ip,1))*sin(0.5*(x(jm,2)+x(jp,2)))
              endif

              sig = 0.5*(kappa_abs(im,jm,km)*rho(im,jm,km) + kappa_abs(ip,jp,kp)*rho(ip,jp,kp)) 

              !FLD 
              ! must use xx with the Er_dif(i,j,k) variable and not T(i,j,k)
              !(if not, some spurious oscillations appear even for equilibrium E=aT4)
              xx = max(1.e-7,abs( (Er_dif(ip,jp,kp) - Er_dif(im,jm,km)) / &
                   (0.5*(Er_dif(ip,jp,kp) + Er_dif(im,jm,km))) ) )
              xx = xx/(0.5*(x(id+1,idim)-x(id-1,idim))*coeff) / sig

              norm = norm + c/sig / (0.5*(xloc(id+1,idim)-xloc(id-1,idim))*coeff) * &
                     lambda_fld(xx) *dt *ds(ip,jp,kp,idim)/dv(i,j,k) 
           enddo
           enddo

           ZZ(i,j,k) = RR(i,j,k) / norm
           
        enddo
     enddo
  enddo

 ! Arrays deallocation
  deallocate(dx,dxc)
  deallocate(ds,dv)
  deallocate(xloc,xcloc)
  deallocate(Tloc)

  if (verbose) write(*,*) 'End of cmp_precond_dif'
  
  return

end subroutine cmp_precond_dif
