!> \file
!! Contains subroutine evol_ray()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine EVOL_RAY
!
!> Performs the evolution over one timestep of the radiative
!! energy and flux in the case of M1 radiative transfer.
!! It computes the evolution using either an explicit
!! method or an implicit method. The simulations will
!! usually always start using the explicit method. Once
!! the implicit timestep is a factor of 'alpha_explicite'
!! (cf. namelist) greater than the explicit timstep, the
!! code switches to implicit resolution.
!<
subroutine evol_ray

  use parameters, only : ndim,nxmin,nxmax
  use divers_ray
  use divers    , only : dt,dtold
  use cputime   , only : temps_comm
  use varray    , only : Eray,Erayold,Fray,Frayold
  use var       , only : Tgaz,E,rhou,rho,B
  use varold    , only : Tgazold
  use para      , only : mype,ncpu

  implicit none

#if PARA==1
  include 'mpif.h'
#else
  real :: MPI_Wtime
#endif

  real    :: xx

  ! Actualization of Tgaz after hydro_step
  !---------------------------------------
  !compute Ein
  E = E - 0.5*sum(rhou**2,4)/rho
  if (allocated(B)) E = E - 0.5*sum(B**2,4)

  call temperature(rho,E,Tgaz,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))

  !restore E 
  E = E + 0.5*sum(rhou**2,4)/rho
  if (allocated(B)) E = E + 0.5*sum(B**2,4)

  call communications_ray(temps_comm)
  call limites_ray


  ! Choice of method
  !-----------------

  if(EXP_RAY) then
  
     Erayold = Eray
     Frayold = Fray
     Tgazold = Tgaz

     if (mype == 0) write(*,*) 'Explicit'
     call explicite

  else

     if (mype == 0) write(*,*) 'Implicit'

     xx = 1.  
     if (dtold .ne. 0.) xx = dt/dtold
     xx = 0.
     
     
     Tgaz = Tgaz + (Tgaz - Tgazold)*xx
     Eray = Eray + (Eray - Erayold)*xx
     Fray = Fray + (Fray - Frayold)*xx
  
     Tgazold = (Tgaz + xx*Tgazold)/(1.+xx)
     Erayold = (Eray + xx*Erayold)/(1.+xx)
     Frayold = (Fray + xx*Frayold)/(1.+xx)
     
     deja_converge =.false.
     
     if(matrix_inv_solver=='GS' .or. matrix_inv_solver=='GS-GMRES') call gs_mat
     
     if((matrix_inv_solver=='GMRES') .or. ((matrix_inv_solver=='GS-GMRES') .and. (deja_converge .eqv. .false.)) &
        .or. (matrix_inv_solver=='GMRES-GS')) then
        call cal_x3d    
        call cal_b3d
        if (mype == 0) write(*,*) 'Inversion GMRES'
        call inversion_gmres
        call calmoinsun_x3d
     endif
     
     if((matrix_inv_solver=='GMRES-GS') .and. (deja_converge .eqv. .false.)) call gs_mat
     
     if(matrix_inv_solver=='yoyo') then
        if (mype == 0) write(*,*) 'Yoyo'
        do while(deja_converge .eqv. .false.)
           if (mype == 0) write(*,*) '  Gauss-Seidel'
           call gs_mat
           call cal_x3d    
           call cal_b3d
           if (mype == 0) write(*,*) '  GMRES'
           call inversion_gmres
           call calmoinsun_x3d
        enddo
     endif

  endif

  ! Reactualization of total energy
  call energy(rho,Tgaz,E,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))
  E = E + 0.5*sum(rhou**2,4)/rho
  if (allocated(B)) E = E + 0.5*sum(B**2,4)
  
  return

end subroutine evol_ray

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine PENTE_MINMOD
!
!> Computes minmod averaged slopes for an arbitrary array
!! of variables.
!<
subroutine pente_minmod(xx,pxx,nxmin,nxmax,nymin,nymax,nzmin,nzmax) 

  use parameters, only : ndim
  use geom      , only : dxc,geometrie,xcloc,geom_dir
!  use divers    , only : pente
  
  implicit none

  integer :: nxmin,nxmax,nymin,nymax,nzmin,nzmax
  integer :: i,j,k
  logical :: pente

  real, dimension(nxmin:nxmax,nymin:nymax,nzmin:nzmax       ) :: xx
  real, dimension(nxmin:nxmax,nymin:nymax,nzmin:nzmax,1:ndim) :: pxx
  real                                                        :: pg,pd

  pxx = 0.
  
  pente = .true.
  
  if(pente) then

     ! slopes along x direction
     do i = min(1,nxmin+1),max(1,nxmax-1)
     do j = min(1,nymin+1),max(1,nymax-1)
     do k = min(1,nzmin+1),max(1,nzmax-1)

        pg = (xx(i  ,j,k) - xx(i-1,j,k))/ dxc(i  ,1)
        pd = (xx(i+1,j,k) - xx(i  ,j,k))/ dxc(i+1,1)
        if (pg*pd <= 0.) then
           pxx(i,j,k,1) = 0.
        else
           pxx(i,j,k,1) = sign(1.,pg)*min(abs(pd),abs(pg))
        endif
        
     enddo
     enddo
     enddo

     ! slopes along y direction
     if(ndim.gt.1) then

        do i=min(1,nxmin+1),max(1,nxmax-1)
        do j=min(1,nymin+1),max(1,nymax-1)
        do k=min(1,nzmin+1),max(1,nzmax-1)

           pg = (xx(i,j  ,k) - xx(i,j-1,k))/ (dxc(j  ,2) )
           pd = (xx(i,j+1,k) - xx(i,j  ,k))/ (dxc(j+1,2) )
           if (pg*pd <= 0.) then
              pxx(i,j,k,2) = 0.
           else
              pxx(i,j,k,2) = sign(1.,pg)*min(abs(pd),abs(pg))
           endif
           
        enddo
        enddo
        enddo
        
     endif

     ! slopes along z direction
     if(ndim.gt.2) then

        do i=min(1,nxmin+1),max(1,nxmax-1)
        do j=min(1,nymin+1),max(1,nymax-1)
        do k=min(1,nzmin+1),max(1,nzmax-1)

           pg = (xx(i,j,k  ) - xx(i,j,k-1))/ dxc(k  ,3)
           pd = (xx(i,j,k+1) - xx(i,j,k  ))/ dxc(k+1,3)
           if (pg*pd <= 0.) then
              pxx(i,j,k,3) = 0.
           else
              pxx(i,j,k,3) = sign(1.,pg)*min(abs(pd),abs(pg))
           endif
           
        enddo
        enddo
        enddo
        
     endif

  endif

  return
  
end subroutine pente_minmod

