!> \file
!! Contains subroutine explicite()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine EXPLICITE
!
!> Performs explicit resolution of one timestep for
!! radiative transfer.
!<
subroutine explicite

!*********************************************************************
!            Different algorithms
!
!  Jacobi       :    D  xn+1 = -(L+U) xn + b
!
!  Gauss-Seidel : (L+D) xn+1 = -   U  xn + b
!*********************************************************************

  use parameters, only : ndim,nxmin,nxmax,nx,nvar_ray,pi
  use divers_ray, only : varrel_ray,alpha,cal_valp,valp_min,rad_trans_model,slope_type_ray
  use divers    , only : mu,dt,verbose,hydro
  use cputime   , only : temps_comm
  use unites    , only : c,c2,seconde
  use geom      , only : x,dx,ds,dv,xcloc,geom_dir,cylindrique,spherique,dxc
  use var_loc   , only : xloc
  use varray    , only : Eray,Fray,Erayold,Frayold,kappa_abs_ray,kappa_emi_ray,kappa_sca_ray
  use var       , only : Tgaz,rho
  use varold    , only : Tgazold
  use para      , only : mype

  implicit none

#if PARA==1
  include 'mpif.h'
  integer :: info,resume_glob
#endif

  integer, dimension(2,3,2,3) :: Iv
  integer, dimension(3)       :: nnx,mx1,mx2
  integer                     :: resume,Iconv
  integer                     :: i,j,k,idim,jdim,kdim,nn
  integer                     :: iface,idp,idm
  integer                     :: ip,jp,kp,im,jm,km
  integer                     :: imin,imax,ipas,iloc

  real, dimension(:,:,:    ), allocatable :: Eray0,Tgaz0
  real, dimension(:,:,:,:  ), allocatable :: Fray0
  real, dimension(:,:,:,:  ), allocatable :: peray,pfray1,pdummy
  real, dimension(:,:,:,:,:), allocatable :: pfray
  real, dimension(1:5,1:5)                :: mat,inv
  real, dimension(1:5    )                :: vect
  real, dimension(     2,ndim)            :: ErayM,ErayP,dEray
  real, dimension(ndim,2,ndim)            :: FrayM,FrayP,dFray
  real, dimension(     2,ndim)            :: ffM, ffP
  real, dimension(1:3,1:3,2,ndim)         :: PrayM,PrayP,DeddP,DeddM
  real, dimension(        2,ndim)         :: eps,ap,am,lm,lp,lmp
  real, dimension(3,3       )             :: Dedd,Dedd_dE
  real, dimension(3,3,1:ndim)             :: Dedd_dF
  real, dimension(3,3,       2,1:ndim)    :: DeddM_dE,DeddP_dE
  real, dimension(3,3,1:ndim,2,1:ndim)    :: DeddM_dF,DeddP_dF
  real, dimension(2,1:ndim)               :: dSloc
  real, dimension(       2,1:ndim)        :: flux_F
  real, dimension(1:ndim,2,1:ndim)        :: flux_P
  real, dimension(     3,2,3     )        :: Rayon
  real                                    :: flux_F_tot
  real, dimension(1:ndim         )        :: flux_P_tot
  real, dimension(       2       )        :: signe
  real                                    :: conv,drp,drm,ff1,ff2
  real                                    :: thetaM,thetaP,interpol_valp
  real                                    :: xx,xm,un_moins,un_plus,f,r,eray_min,tgaz_min
  real                                    :: source_ray,derive_source_ray

  logical :: P_old,chi_uniforme,derive_chi

!*********************************************************************************************************************************
!  Dedd,Dedd_dE (i,j)
!  Dedd_dF      (i,j,k)
!
!  DeddM_dE, DeddP_dE  (i,j  ,iface,idim)
!  DeddM_dF, DeddP_dF  (i,j,k,iface,idim)
!  
!  Dedd_1d,Dedd_dE_1d,Dedd_2d,Dedd_dE_2d,Dedd_3d,Dedd_dE_3d  (i,j  ,iface)
!  Dedd_dF_1d,Dedd_dF_2d,Dedd_dF_3d                          (i,j,k,iface)
!
!   iface : face index            (=1:   2) with 1 = back ; 2 = front
!   idim  : direction considered  (=1:ndim)
!
!   i,j   : tensorial coordinates       (=1:3   )
!   k     : flux direction considered   (=1:ndim)
!
!*********************************************************************************************************************************
  
  if (verbose) write(*,*) 'Start of explicite'
  
  ! In order to have a uniform chi in the cells
  Chi_uniforme = .false.

  ! In order to include the derivatives of chi in the matrix
  Derive_Chi   = .true.

  ! incoming and outgoing fluxes
  signe(1) = -1. ; signe(2) = 1.

   do i = 1,3
      nnx(i) = 4 + nx(i)
      mx1(i) = max(-1,nxmin(i))
      mx2(i) = min(nnx(i)-2,nxmax(i))
   enddo
   
  ! geometry
  nn = max(nx(1),nx(2),nx(3))
  allocate(xloc(-1:nn+3,ndim))
  
                xloc(-1:nx(1)+3,1) = x(-1:nx(1)+3,1)   
 if (ndim.gt.1) xloc(-1:nx(2)+3,2) = x(-1:nx(2)+3,2)   
 if (ndim.gt.2) xloc(-1:nx(3)+3,3) = x(-1:nx(3)+3,3)   

  call cal_geom(nx(1),nx(2),nx(3))

  ! neighbour offsets
  ! Iv(i,j,k,l) :: i = front or back face -- j = face direction -- k = front or back cell -- l = cell coordinate
  Iv = 0
  do idim = 1,3
     Iv(1,idim,1,idim) = -1
     Iv(2,idim,2,idim) = +1
  enddo

  allocate(Tgaz0 (mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3)              ))
  allocate(Eray0 (mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3)              ))
  allocate(Fray0 (mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3),1:ndim       ))
  
  allocate(peray (mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3),1:ndim       ))
  allocate(pfray1(mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3),1:ndim       ))
  allocate(pdummy(mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3),1:ndim       ))
  allocate(pfray (mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3),1:ndim,1:ndim)) 

  P_old = .false.

  un_moins = 1. !- 1.0E-8
  un_plus  = 1. !+ 1.0E-8
  
  eray_min = 4.0e-16
  tgaz_min = 0.5

  ! Boundary conditions
  call communications_ray(temps_comm)
  call limites_ray

  call cal_kappa_ray(rho,Tgaz,kappa_emi_ray,kappa_sca_ray,nxmin(1),nxmax(1),nxmin(2),nxmax(2),nxmin(3),nxmax(3))
  kappa_abs_ray = kappa_emi_ray

  do k = nxmin(3),nxmax(3)
     do j = nxmin(2),nxmax(2)
        do i = nxmin(1),nxmax(1)
           if (Eray(i,j,k) <= 0.) then
              write(*,*) '** cal_ray **'
              write(*,*) 'i,j,k,Eray(i,j,k)= ',i,j,k,Eray(i,j,k)
              Eray(i,j,k) = eray_min
!               read(*,*)
           endif
        enddo
     enddo
  enddo

  do k = nxmin(3),nxmax(3)
     do j = nxmin(2),nxmax(2)
        do i = nxmin(1),nxmax(1)
           if ( Tgaz(i,j,k) < 0.) then
              write(*,*) 'Debut de Ray'
              write(*,*) 'Tgaz =', Tgaz(i,j,k)
              write(*,*) 'i,j,k=',i,j,k
              Tgaz(i,j,k) = tgaz_min
!               read(*,*)
           endif
        enddo
     enddo
  enddo

  do k = nxmin(3),nxmax(3)
     do j = nxmin(2),nxmax(2)
        do i = nxmin(1),nxmax(1)
           f = 0.0
           do idim = 1,ndim
              f = f + (Fray(i,j,k,idim)/(c*Eray(i,j,k)))**2
           enddo
           f = sqrt(f)
           if(f > 1.0)then
              write(*,*) 'F/cE > 1 : i,j,k,F/cE,fx,fy,fz =',i,j,k,f,(Fray(i,j,k,idim)/(c*Eray(i,j,k)),idim=1,ndim)
              r = 1.0 / f
              do idim = 1,ndim
                 Fray(i,j,k,idim) = Fray(i,j,k,idim) * r
              enddo
!               read(*,*)
           endif
        enddo
     enddo
  enddo

  do k = mx1(3),mx2(3)
  do j = mx1(2),mx2(2)
  do i = mx1(1),mx2(1)
  
     Tgaz0(i,j,k       ) = Tgazold(i,j,k       )
     Eray0(i,j,k       ) = Erayold(i,j,k       )
     Fray0(i,j,k,1:ndim) = Frayold(i,j,k,1:ndim)

  enddo
  enddo
  enddo

  if (verbose) write(*,*) '     -- Eray slope --'
  call slopes(Eray0,peray,pdummy,mx1(1),mx2(1),mx1(2),mx2(2),mx1(3),mx2(3),slope_type_ray)
  if (verbose) write(*,*) '     -- Fray slope --'
  do idim = 1,ndim
     call slopes(Fray0(mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3),idim),pfray1,pdummy, &
                 mx1(1),mx2(1),mx1(2),mx2(2),mx1(3),mx2(3),slope_type_ray)
     pfray(mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3),idim,1:ndim) = pfray1(mx1(1):mx2(1),mx1(2):mx2(2),mx1(3):mx2(3),1:ndim)
  enddo

222 resume = 0

  do k = 1,nx(3)
  do j = 1,nx(2)
  do i = 1,nx(1)

     PrayM = 0. ; PrayP = 0.

     if (verbose) write(*,*) '     -- Interface values --'

     do idim  = 1,ndim
     do iface = 1,2

        im = i + Iv(iface,idim,1,1) ; jm = j + Iv(iface,idim,1,2); km = k + Iv(iface,idim,1,3)
        ip = i + Iv(iface,idim,2,1) ; jp = j + Iv(iface,idim,2,2); kp = k + Iv(iface,idim,2,3)

        if(idim.eq.1) then
           idm = im
           idp = ip
        else if(idim.eq.2) then
           idm = jm
           idp = jp
        else if(idim.eq.3) then
           idm = km
           idp = kp
        endif

        drp = 0.5*(x(idp  ,idim)-x(idm,idim))
        drm = 0.5*(x(idp+1,idim)-x(idp,idim)) 

        ErayM(iface,idim) = Eray0(im,jm,km) + drp*peray(im,jm,km,idim)
        ErayP(iface,idim) = Eray0(ip,jp,kp) - drm*peray(ip,jp,kp,idim)

        if(iface.eq.1) then
           dEray(iface,idim) = ErayP(iface,idim) - Eray0(i,j,k) 
        else
           dEray(iface,idim) = ErayM(iface,idim) - Eray0(i,j,k) 
        endif

        do jdim = 1,ndim
           FrayM(jdim,iface,idim) = Fray0(im,jm,km,jdim) + drp*pfray(im,jm,km,jdim,idim)
           FrayP(jdim,iface,idim) = Fray0(ip,jp,kp,jdim) - drm*pfray(ip,jp,kp,jdim,idim)
        enddo

        ! Ensure that f < 1
        f = 0.0
        do jdim = 1,ndim
           f = f + (FrayM(jdim,iface,idim)/(c*ErayM(iface,idim)))**2
        enddo
        f = sqrt(f)
        if(f > 1.0)then
           r = 1.0 / f
           do jdim = 1,ndim
              FrayM(jdim,iface,idim) = FrayM(jdim,iface,idim) * r
           enddo
        endif
        f = 0.0
        do jdim = 1,ndim
           f = f + (FrayP(jdim,iface,idim)/(c*ErayP(iface,idim)))**2
        enddo
        f = sqrt(f)
        if(f > 1.0)then
           r = 1.0 / f
           do jdim = 1,ndim
              FrayP(jdim,iface,idim) = FrayP(jdim,iface,idim) * r
           enddo
        endif

        ffM(iface,idim) = 0. ; ffP(iface,idim) = 0.
        do jdim = 1,ndim
           if(iface.eq.1) then
              dFray(jdim,iface,idim) = FrayP(jdim,iface,idim)- Fray0(i,j,k,jdim)
           else
              dFray(jdim,iface,idim) = FrayM(jdim,iface,idim)- Fray0(i,j,k,jdim)
           endif

           ffM(iface,idim) = ffM(iface,idim) + FrayM(jdim,iface,idim)**2
           ffP(iface,idim) = ffP(iface,idim) + FrayP(jdim,iface,idim)**2
        enddo
        ffM(iface,idim) = sqrt(ffM(iface,idim))/(c*ErayM(iface,idim))
        ffP(iface,idim) = sqrt(ffP(iface,idim))/(c*ErayP(iface,idim))

        ! Calculate theta incidence angle to interpolate eigenvalues
        ! theta needs to be calculated here before ffM is changed if it is greater than 1
        if(ffM(iface,idim).gt.1.e-5) then
           xx=FrayM(idim,iface,idim)/(c*ErayM(iface,idim))/ffM(iface,idim)
           if(abs(xx).gt.1.) then
              thetaM=pi/2.*(1.-sign(1.,xx))
           else
              thetaM=acos(xx)
           endif
        else
           thetaM=0.
        endif
        if(ffP(iface,idim).gt.1.e-5) then
           xx=FrayP(idim,iface,idim)/(c*ErayP(iface,idim))/ffP(iface,idim)
           if(abs(xx).gt.1.) then
              thetaP=pi/2.*(1.-sign(1.,xx))
           else
              thetaP=acos(xx)
           endif
        else
           thetaP=0.
        endif

        ! If ffM > 1 then set it to 1 - WARNING: NOT VERY SATISFACTORY!! Ed
        if (ffM(iface,idim).gt. 1.) ffM(iface,idim)= 1.
        if (ffP(iface,idim).gt. 1.) ffP(iface,idim)= 1.

        ! Compute Eddington tensor and its derivatives at the cell centre and interfaces
        if (Chi_uniforme) then 
           call cal_Dedd(Eray0(ip,jp,kp),Fray0(ip,jp,kp,1:ndim),DeddP(:,:,iface,idim),DeddP_dE(:,:,iface,idim),DeddP_dF(:,:,:,iface,idim) , Derive_chi)
           call cal_Dedd(Eray0(im,jm,km),Fray0(im,jm,km,1:ndim),DeddM(:,:,iface,idim),DeddM_dE(:,:,iface,idim),DeddM_dF(:,:,:,iface,idim) , Derive_chi)
        else
           call cal_Dedd(ErayP(iface,idim),FrayP(1:ndim,iface,idim),DeddP(:,:,iface,idim),DeddP_dE(:,:,iface,idim),DeddP_dF(:,:,:,iface,idim) , Derive_chi)
           call cal_Dedd(ErayM(iface,idim),FrayM(1:ndim,iface,idim),DeddM(:,:,iface,idim),DeddM_dE(:,:,iface,idim),DeddM_dF(:,:,:,iface,idim) , Derive_chi)
        endif

        do jdim=1,ndim
        do kdim=1,ndim
           PrayM(jdim,kdim,iface,idim) = DeddM(jdim,kdim,iface,idim)*ErayM(iface,idim)
           PrayP(jdim,kdim,iface,idim) = DeddP(jdim,kdim,iface,idim)*ErayP(iface,idim)
        enddo
        enddo

        ! Compute epsilon for asymptotic preserving scheme
        if (verbose) write(*,*) '     -- compute epsilon --'

        ff1 = ffM(iface,idim)
        ff2 = ffP(iface,idim)

!!$           if(kappa_abs_ray(im,jm,km)*rho(im,jm,km).ne.0. .and. kappa_abs_ray(ip,jp,kp)*rho(ip,jp,kp).ne.0.) then
!!$              eps(iface,idim) = max( 1./(kappa_abs_ray(im,jm,km)*rho(im,jm,km)*(2.*drm)) &
!!$                   &                ,1./(kappa_abs_ray(ip,jp,kp)*rho(ip,jp,kp)*(2.*drp)) ) 
        if( ((kappa_abs_ray(im,jm,km)+kappa_sca_ray(im,jm,km))*rho(im,jm,km)).ne.0. .and. &
            ((kappa_abs_ray(ip,jp,kp)+kappa_sca_ray(ip,jp,kp))*rho(ip,jp,kp)).ne.0.       ) then
           eps(iface,idim) = max( 1./((kappa_abs_ray(im,jm,km)+kappa_sca_ray(im,jm,km))*rho(im,jm,km)*(2.*drm)) &
                &                ,1./((kappa_abs_ray(ip,jp,kp)+kappa_sca_ray(ip,jp,kp))*rho(ip,jp,kp)*(2.*drp)) )
           eps(iface,idim) = max(eps(iface,idim),2.*abs(ff1),2.*abs(ff2)) 
           eps(iface,idim) = min(1.,eps(iface,idim)) 
        else
           eps(iface,idim) = 1.
        endif

        if (verbose) write(*,*) '     -- compute eigenvalues --'

        ! Compute eigenvalues as a fonction of normef, theta and epsilon
        if (cal_valp) then

           select case(rad_trans_model)
   
           case('P1')
   
              am(iface,idim)=min(-valp_min,-1./sqrt(3.))
              ap(iface,idim)=max(+valp_min, 1./sqrt(3.))
      
           case('M1')
   
              am(iface,idim)=min(-valp_min,interpol_valp(ffM(iface,idim),thetaM,eps(iface,idim),1), &
                                           interpol_valp(ffP(iface,idim),thetaP,eps(iface,idim),1))
              ap(iface,idim)=max(+valp_min,interpol_valp(ffM(iface,idim),thetaM,eps(iface,idim),4), &
                                           interpol_valp(ffP(iface,idim),thetaP,eps(iface,idim),4))
   
           end select
   
        else
           am(iface,idim) = -1.
           ap(iface,idim) =  1.
        endif
        
        if(ap(iface,idim).ne.am(iface,idim)) then
           lm (iface,idim) = am(iface,idim)               /(ap(iface,idim)-am(iface,idim))
           lp (iface,idim) =                ap(iface,idim)/(ap(iface,idim)-am(iface,idim))
           lmp(iface,idim) = am(iface,idim)*ap(iface,idim)/(ap(iface,idim)-am(iface,idim))
        else ! then assume ap=am=0.
           lm (iface,idim) = -1./2
           lp (iface,idim) =  1./2
           lmp(iface,idim) =  0.
        endif

     enddo
     enddo

     call cal_Dedd(Eray0(i,j,k),Fray0(i,j,k,1:ndim),Dedd,Dedd_dE,Dedd_dF,Derive_chi)

     ! Initialise the geometry
     do idim = 1,ndim
        ip = i ; jp = j ; kp = k
        if(idim==1) ip = i+1
        if(idim==2) jp = j+1
        if(idim==3) kp = k+1
        dSloc(1,idim) = dS(i ,j ,k ,idim)
        dSloc(2,idim) = dS(ip,jp,kp,idim)
     enddo


     !******************************************************************
     !                     Compute fluxes
     !******************************************************************
     rayon = 1.
     if (geom_dir(2) == 'phi_c' .or. geom_dir(2) == 'the_s') then
        rayon(2,1:2,1:3) = xcloc(i  ,1)
        rayon(2,1  ,1  ) =  xloc(i  ,1)
        rayon(2,2  ,1  ) =  xloc(i+1,1)
     endif

     if (geom_dir(3) == 'phi_s' ) then
        rayon(3,1  ,1  ) =  xloc(i  ,1) * sin(xcloc(j  ,2))
        rayon(3,2  ,1  ) =  xloc(i+1,1) * sin(xcloc(j  ,2))
        rayon(3,1  ,2  ) = xcloc(i  ,1) * sin( xloc(j  ,2))
        rayon(3,2  ,2  ) = xcloc(i  ,1) * sin( xloc(j+1,2))
        rayon(3,1  ,3  ) = xcloc(i  ,1) * sin(xcloc(j  ,2))
        rayon(3,2  ,3  ) = xcloc(i  ,1) * sin(xcloc(j  ,2))
     endif

     flux_F = 0.
     flux_P = 0.
     do idim=1,ndim
        flux_F(1,idim) = (+  lp (1,idim)            *FrayM(idim,1,idim)                 &
                          -  lm (1,idim)            *FrayP(idim,1,idim)                 &
                          -c*lmp(1,idim)*eps(1,idim)*ErayM(     1,idim)                 &
                          +c*lmp(1,idim)*eps(1,idim)*ErayP(     1,idim) )*dSloc(1,idim)
        flux_F(2,idim) = (-  lm (2,idim)            *FrayP(idim,2,idim)                 &
                          +  lp (2,idim)            *FrayM(idim,2,idim)                 &
                          +c*lmp(2,idim)*eps(2,idim)*ErayP(     2,idim)                 &
                          -c*lmp(2,idim)*eps(2,idim)*ErayM(     2,idim) )*dSloc(2,idim)
     enddo

     do jdim = 1,ndim
        imin=1
        imax=ndim
        ipas=1
        if (jdim.eq.1 .and. (geom_dir(1) == 'rc' .or. geom_dir(1) == 'rs') ) then
           imin=2
           idim=1
           flux_P(jdim,1,idim) = (  c2*( lp (1,idim)*eps(1,idim)**2+(1.-eps(1,idim)**2)/2)* PrayM(jdim,idim,1,idim) &
                                   -c *  lmp(1,idim)*eps(1,idim)                          * FrayM(jdim     ,1,idim) &
                                   +c *  lmp(1,idim)*eps(1,idim)                          * FrayP(jdim     ,1,idim) &
                                   +c2*(-lm (1,idim)*eps(1,idim)**2+(1.-eps(1,idim)**2)/2)* PrayP(jdim,idim,1,idim) ) *dV(i,j,k)/(xloc(i+1,1)-xloc(i,1))
           flux_P(jdim,2,idim) = (  c2*(-lm (2,idim)*eps(2,idim)**2+(1.-eps(2,idim)**2)/2)* PrayP(jdim,idim,2,idim) &
                                   +c *  lmp(2,idim)*eps(2,idim)                          * FrayP(jdim     ,2,idim) &
                                   -c *  lmp(2,idim)*eps(2,idim)                          * FrayM(jdim     ,2,idim) &
                                   +c2*( lp (2,idim)*eps(2,idim)**2+(1.-eps(2,idim)**2)/2)* PrayM(jdim,idim,2,idim) ) *dV(i,j,k)/(xloc(i+1,1)-xloc(i,1))
        endif
        if (jdim.eq.2 .and. geom_dir(2) == 'the_s') then
           ipas=2
           idim=2
           flux_P(jdim,1,idim) = (  c2*( lp (1,idim)*eps(1,idim)**2+(1.-eps(1,idim)**2)/2)* PrayM(jdim,idim,1,idim) &
                                   -c *  lmp(1,idim)*eps(1,idim)                          * FrayM(jdim     ,1,idim) &
                                   +c *  lmp(1,idim)*eps(1,idim)                          * FrayP(jdim     ,1,idim) &
                                   +c2*(-lm (1,idim)*eps(1,idim)**2+(1.-eps(1,idim)**2)/2)* PrayP(jdim,idim,1,idim) ) *dV(i,j,k)/(xloc(j+1,2)-xloc(j,2)) !/xcloc(i,1)
           flux_P(jdim,2,idim) = (  c2*(-lm (2,idim)*eps(2,idim)**2+(1.-eps(2,idim)**2)/2)* PrayP(jdim,idim,2,idim) &
                                   +c *  lmp(2,idim)*eps(2,idim)                          * FrayP(jdim     ,2,idim) &
                                   -c *  lmp(2,idim)*eps(2,idim)                          * FrayM(jdim     ,2,idim) &
                                   +c2*( lp (2,idim)*eps(2,idim)**2+(1.-eps(2,idim)**2)/2)* PrayM(jdim,idim,2,idim) ) *dV(i,j,k)/(xloc(j+1,2)-xloc(j,2)) !/xcloc(i,1)
        endif
        do idim=imin,imax,ipas
           flux_P(jdim,1,idim) = (  c2*( lp (1,idim)*eps(1,idim)**2+(1.-eps(1,idim)**2)/2)* PrayM(jdim,idim,1,idim) &
                                   -c *  lmp(1,idim)*eps(1,idim)                          * FrayM(jdim     ,1,idim) &
                                   +c *  lmp(1,idim)*eps(1,idim)                          * FrayP(jdim     ,1,idim) &
                                   +c2*(-lm (1,idim)*eps(1,idim)**2+(1.-eps(1,idim)**2)/2)* PrayP(jdim,idim,1,idim) ) *dSloc(1,idim) *rayon(jdim,1,idim)
           flux_P(jdim,2,idim) = (  c2*(-lm (2,idim)*eps(2,idim)**2+(1.-eps(2,idim)**2)/2)* PrayP(jdim,idim,2,idim) &
                                   +c *  lmp(2,idim)*eps(2,idim)                          * FrayP(jdim     ,2,idim) &
                                   -c *  lmp(2,idim)*eps(2,idim)                          * FrayM(jdim     ,2,idim) &
                                   +c2*( lp (2,idim)*eps(2,idim)**2+(1.-eps(2,idim)**2)/2)* PrayM(jdim,idim,2,idim) ) *dSloc(2,idim) *rayon(jdim,2,idim)
        enddo
     enddo

     flux_F_tot = 0.
     flux_P_tot = 0.
     do idim=1,ndim
        do iface=1,2
           flux_F_tot = flux_F_tot + signe(iface)*dt*flux_F(iface,idim)
           do jdim=1,ndim
              flux_P_tot(jdim) = flux_P_tot(jdim) + signe(iface)*dt*flux_P(jdim,iface,idim) !*rayon(jdim,iface,idim)
           enddo
        enddo
     enddo
     
     flux_F_tot = flux_F_tot/dV(i,j,k)
     do idim=1,ndim
        flux_P_tot(idim) = flux_P_tot(idim)/dV(i,j,k)/rayon(idim,2,3)
     enddo

     if (cylindrique) then
        iloc=3
        if(geom_dir(2)=='phi_c') iloc=2
        flux_P_tot(1) = flux_P_tot(1) + c2*dt/dV(i,j,k)* ( Dedd(1,1)-Dedd(iloc,iloc) ) * Eray0(i,j,k) *(dSloc(2,1)-dSloc(1,1))
     endif
     if (spherique) then
                                 flux_P_tot(1) = flux_P_tot(1) + c2*dt/dV(i,j,k)*                                 &
             (2.*Dedd(1,1)-Dedd(2,2)-Dedd(3,3) ) * Eray0(i,j,k) *0.5*(dSloc(2,1)-dSloc(1,1))
        if(geom_dir(2)=='the_s') flux_P_tot(2) = flux_P_tot(2) + c2*dt/dV(i,j,k)* cos(xcloc(j,2))/sin(xcloc(j,2))*&
             (             Dedd(2,2)-Dedd(3,3) ) * Eray0(i,j,k) *0.5*(dSloc(2,1)-dSloc(1,1))
     endif
        

     ! Initialise mat
     vect = 0.
     mat  = 0.
     do idim=1,5
        mat(idim,idim) = 1.
     enddo

     ! Er equation with no source term
     mat(2,1) = 0.
     mat(2,2) = 1. 
     
     vect(2) = vect(2) - Erayold(i,j,k) + flux_F_tot

     ! Tgaz equation (using the coefficients from the Er equation without the source term)
     call heat_capacity(rho(i,j,k),Tgaz0(i,j,k),conv,1,1,1,1,1,1)
     conv=1./conv

     mat(1,1  ) = 1.
     mat(1,2:5) = mat(2,2:5)* conv 
     
     vect(1)    = vect(2)*conv -Tgazold(i,j,k)

     ! Finally include the energy source term: rho c (kappa_em aT^4-kappa_abs E)
     
     !  T4 = Tgazold(i,j,k)**4 *(1. + 4.*(Tgaz(i,j,k)-tgazold(i,j,k))/Tgazold(i,j,k)) 
     !  T4 = -3.*Tgazold(i,j,k)**4 + 4.* Tgazold(i,j,k)**3*Tgaz(i,j,k)
     
     mat(2,1) =          - dt*kappa_emi_ray(i,j,k)*rho(i,j,k)*c*derive_source_ray(Tgaz0(i,j,k))
     mat(2,2) = mat(2,2) + dt*kappa_abs_ray(i,j,k)*rho(i,j,k)*c
     vect(2)  = vect(2)  + dt*kappa_emi_ray(i,j,k)*rho(i,j,k)*c*(derive_source_ray(Tgaz0(i,j,k))*Tgaz0(i,j,k)-source_ray(Tgaz0(i,j,k)))
  
     ! Fr equation using only dtFr + c2(Div . Pr) = 0 terms
     do idim = 1,ndim
     
        mat(2+idim,1     ) = 0.
        mat(2+idim,2     ) = 0.
        mat(2+idim,2+idim) = 1.

        vect(2+idim) = -Frayold(i,j,k,idim) + vect(2+idim) + flux_P_tot(idim)

        ! Include the flux source term: -kappa_abs rho c F
        mat(2+idim,2+idim) = mat(2+idim,2+idim) + c*dt*(kappa_abs_ray(i,j,k)+kappa_sca_ray(i,j,k))*rho(i,j,k)

     enddo


     !**************  Normalised equations   ****************!
     !*******************************************************!
     !**        eqi / Xoldi et Aij=dFi/dXj *Xoldj          **!
     !**       to get     Aij (deltaX/Xold)j = eqi         **!
     !*******************************************************!
     !*******************************************************!
     vect(1  ) = vect(1  )/   Tgazold(i,j,k)
     vect(2  ) = vect(2  )/   Erayold(i,j,k)
     vect(3:5) = vect(3:5)/(c*Erayold(i,j,k))

     mat(1  ,1:5) = mat(1  ,1:5)/   Tgazold(i,j,k)
     mat(2  ,1:5) = mat(2  ,1:5)/   Erayold(i,j,k)
     mat(3:5,1:5) = mat(3:5,1:5)/(c*Erayold(i,j,k))

     mat(1:5,1  ) = mat(1:5,1  ) *   Tgazold(i,j,k)
     mat(1:5,2  ) = mat(1:5,2  ) *   Erayold(i,j,k)
     mat(1:5,3:5) = mat(1:5,3:5) * c*Erayold(i,j,k)


     vect(1:5) = - vect(1:5)


     call essai_inv5x5(mat,inv)
     vect(1:nvar_ray           ) = matmul( inv(1:nvar_ray,1:nvar_ray),vect(1:nvar_ray           ) )
     mat (1:nvar_ray,1:nvar_ray) = matmul( inv(1:nvar_ray,1:nvar_ray),mat (1:nvar_ray,1:nvar_ray) )


     Tgaz(i,j,k     ) = vect(1     ) *    Tgazold(i,j,k) 
     Eray(i,j,k     ) = vect(2     ) *    Erayold(i,j,k)
     do idim=1,ndim
        Fray(i,j,k,idim) = vect(2+idim) *  c*Erayold(i,j,k)
     enddo

  enddo   ! End of i,j,k loops
  enddo
  enddo

        
  do k = 1,nx(3)
  do j = 1,nx(2)
  do i = 1,nx(1)

     f = 0.0
     do idim = 1,ndim
        f = f + (Fray(i,j,k,idim)/(c*Eray(i,j,k)))**2
     enddo
     f = sqrt(f)
     if(f > 1.0)then
        write(*,*) 'F/cE > 1 : i,j,k,F/cE,fx,fy,fz =',i,j,k,f,(Fray(i,j,k,idim)/(c*Eray(i,j,k)),idim=1,ndim)
        r = 1.0 / f
        do idim = 1,ndim
           Fray(i,j,k,idim) = Fray(i,j,k,idim) * r
        enddo
!         read(*,*)
     endif

  enddo
  enddo
  enddo

  ! Boundary conditions
  call communications_ray(temps_comm)
  call limites_ray


#if PARA==1
  ! Resume reduction to see if f > 1 somewhere in the grid and if so then reduce dt
  info = 0
  call MPI_ALLREDUCE(resume,resume_glob,1,MPI_INTEGER,MPI_SUM,MPI_COMM_WORLD,info)
  resume = resume_glob
#endif
  
  if(resume .gt. 0) then
     if(mype==0) write(*,*) 'Explicite : Resume=',resume
     dt = dt * 0.8
     if(mype==0) write(*,*) 'dt =',dt/seconde,' s'
     !read(*,*)
     goto 222
  endif

  if (verbose) write(*,*) 'End of explicite'

  if(mype==0) write(*,*) ''

  ! Array deallocation
  deallocate(Tgaz0,Eray0,Fray0)
  deallocate(dx,dxc)
  deallocate(ds,dv)
  deallocate(xloc,xcloc)
  deallocate(peray,pfray1,pfray,pdummy)
  
  return

end subroutine explicite
