!> \file
!! Contains subroutine gmres()
!<

!###########################################################
!###########################################################
!###########################################################
!###########################################################

!  Subroutine GMRES
!
!> Performs gmres implicit resolution of one timestep for
!! radiative transfer.
!!
!! n      : total problem size
!! m      : restart
!! nloc   : local problem size
!! work (cf doc gmres) : workspace array for gmres routine
!!                       work(1     :  nloc) : solution or initial estimate
!!                       work(1+nloc:2*nloc) : second term
!!
!! workred : workspace array with overlap zones
!! lwork   : size of work
!! cntl, icntl : parameter arrays for gmres (cf doc gmres)
!!
!<
subroutine gmres(n,m,nloc,work,nx,ny,nz,nvar_ray,imin,imax,jmin,jmax, &
                 kmin,kmax,workred,lwork,cntl,icntl,info)
		 
  use para, only : mype

  implicit none
  
#if PARA==1
  include 'mpif.h'
  integer :: ierr
#endif

  ! input arguments
  integer, intent(in) :: n,m,nloc,lwork,nx,ny,nz,nvar_ray,imin,imax,jmin,jmax,kmin,kmax
  real, dimension(-nvar_ray+1:(imax-imin+(jmax-jmin)*(imax-imin+1)+&
       (kmax-kmin)*(imax-imin+1)*(jmax-jmin+1))*nvar_ray) :: workred
  real, dimension(lwork) :: work
  real, dimension(5)     :: cntl
  integer, dimension(7)  :: icntl
  integer,dimension(3)   :: info

  ! local variables
  integer              :: revcom,colx,coly,colz,nbscal
  integer,dimension(5) :: irc(5)
  real, dimension(2)   :: rinfo =1.d0
  integer,parameter    :: matvec=1, precondLeft=2, precondRight=3, dotProd=4
  real,parameter       :: ZERO = 0.0d0, ONE = 1.0d0
  integer              :: i,j,k,ivar


!****************************************
!* Reverse communication implementation
!****************************************

10 call drive_dgmres(n,nloc,m,lwork,work,irc,icntl,cntl,info,rinfo)

  revcom = irc(1)
  colx   = irc(2)
  coly   = irc(3)
  colz   = irc(4)
  nbscal = irc(5)

  if (revcom.eq.matvec) then
     ! perform the matrix vector product
     !     work(colz) <-- A * work(colx)
     do ivar = 1,nvar_ray
        do k = 1,nz
           do j = 1,ny
              do i = 1,nx
                 workred((i-imin+(j-jmin)*(imax-imin+1)+(k-kmin)*(imax-imin+1)*(jmax-jmin+1)-1)*nvar_ray+ivar) = &
                         work(colx-1+(i+(j-1)*nx+(k-1)*nx*ny-1)*nvar_ray+ivar)
              enddo
           enddo
        enddo
     enddo

     call mat_prod3d(workred,work(colz))
 
     goto 10

  else if (revcom.eq.precondLeft) then
     ! perform the left preconditioning
     !         work(colz) <-- M^{-1} * work(colx)
     call close_program(mype,'Left Preconditionning GMRES not implemented')
     goto 10

  else if (revcom.eq.precondRight) then
     ! perform the right preconditioning
     call close_program(mype,'Right Preconditionning GMRES not implemented')
     goto 10

  else if (revcom.eq.dotProd) then
     ! perform the scalar product
     ! work(colz) <-- work(colx) work(coly)
     ! where work(colx) = n vectors  
     call dgemv('T',nloc,nbscal,ONE,work(colx),nloc, work(coly),1,ZERO,work(colz),1)
#if PARA==1
     call MPI_Allreduce(work(colz),workred(1),nbscal,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,ierr)
#else
     workred(1:nbscal)=work(colz:colz+nbscal-1)
#endif
     work(colz:colz+nbscal-1) = workred(1:nbscal)
     goto 10
  endif
  
  return

end subroutine gmres
