;+
; NAME:
;       RD
;
; PURPOSE:
;       This procedure reads a HERACLES output file (.Bin format)
;
; CATEGORY:
;       Input/Output
;
; CALLING SEQUENCE:
;       RD, nout, var, verbose=verbose
;
; INPUTS:
;       NOUT:   output number
;
; OPTIONAL INPUTS:
;       VERBOSE: to display some informations
;
;       DIR: directory where the files are stored (defaut: current directory)
;
; OUTPUTS:
;       var:  Store all the variables in structure var
;
; EXAMPLE:
;      rd,1,var
;-
pro rd, nout, var, verbose=verbose, dir=dir

  IF N_PARAMS() NE 2 THEN BEGIN
     PRINT, 'Wrong number of arguments'
     DOC_LIBRARY,'rd'
     RETURN
  ENDIF
  
  ;Reading the first file only to get
  ;the dimensions of the simulation and
  ;to allocate arrays
  ;************************************
  if (nout lt 10) then begin
     out = '00'+strcompress(string(nout),/remove_all)
  endif else if (nout lt 100) then begin
     out = '0'+strcompress(string(nout),/remove_all)
  endif else begin
     out = ''+strcompress(string(nout),/remove_all)
  endelse
 
  if not keyword_set(dir) then dir='.'
  dir=dir+'/'+out
  out='_'+out

  name = 'liste-00001'+out+'.Bin'
  name = dir+'/'+name

  openr,1,name,/f77_unformatted
  nxg =1L & nyg=1L & nzg =1L
  ncpu=1L & geometry='x'
  ndim=1L 
  nx  =1L & ny =1L & nz  =1L
  nfx = 1L
  nbuf= 1L
  ngrp_ray= 0L
  ngrp_dif= 0L

  mhd=1L & rhd=1L & dif=1L & cond=1L & grav=1L

  xg  = dblarr(3)
  x1  = 1D0
  x2  = 1D0
  x3  = 1D0
  
  readu,1,nxg,nyg,nzg,ncpu,geometry
  xg=dblarr(1,nxg+1) &   yg=dblarr(1,nyg+1) &   zg=dblarr(1,nzg+1)  
  readu,1, xg,yg,zg
  readu,1,ndim,nx,ny,nz,nfx,nbuf,ngrp_ray,ngrp_dif
  readu,1,mhd,rhd,dif,cond,grav
  close,1

  nmax = max([nxg,nyg,nzg])
  xx = dblarr(3,nmax+1)
  xx(0,0:nxg) = xg
  xx(1,0:nyg) = yg
  xx(2,0:nzg) = zg

  ;Fill the mesh structure
  ;***********************
  mesh={ndim:ndim,nx:nxg,ny:nyg,nz:nzg,nfx:nfx,x:fltarr(nxg+1),y:fltarr(nyg+1),z:fltarr(nzg+1),xc:fltarr(nxg),yc:fltarr(nyg),zc:fltarr(nzg),geometry:geometry}
  mesh.x=xg(0,0:nxg)
  mesh.y=yg(0,0:nyg)
  mesh.z=zg(0,0:nzg)
  for i=0,nxg-1 do begin
     mesh.xc(i)=0.5*(xg(0,i)+xg(0,i+1))
  endfor
  for j=0,nyg-1 do begin
     mesh.yc(j)=0.5*(yg(0,j)+yg(0,j+1))
  endfor
  for k=0,nzg-1 do begin
     mesh.zc(k)=0.5*(zg(0,k)+zg(0,k+1))
  endfor

  ;The variable structure
  ;**********************
;   shift_gr= dblarr(ndim)
;  nshift_gr= lonarr(ndim)
  k = 1L
  time = 1D0 & dt = 1D0 & dtold = 1D0 
  gamma = 1D0 & mu = 1D0
  u_M = 1D0 & u_L = 1D0 & u_T = 1D0
  var={time:time,dt:dt,dtold:dtold,nite:k,mesh:mesh,mu:mu,gamma:gamma,u_M:u_M,u_L:u_L,u_T:u_T,rho:fltarr(nxg,nyg,nzg),u:fltarr(nxg,nyg,nzg,ndim),E:fltarr(nxg,nyg,nzg),p:fltarr(nxg,nyg,nzg),eint:fltarr(nxg,nyg,nzg),T:fltarr(nxg,nyg,nzg)}
  if(nfx ge 1) then var=create_struct(var,'fx',fltarr(nxg,nyg,nzg,nfx))
  if mhd then var=create_struct(var,'B',fltarr(nxg,nyg,nzg,3))
  if rhd then var=create_struct(var,'ngrp_ray',ngrp_ray,'er',fltarr(nxg,nyg,nzg,ngrp_ray),'fr',fltarr(nxg,nyg,nzg,ndim,ngrp_ray))
  if dif then var=create_struct(var,'ngrp_dif',ngrp_dif,'ed',fltarr(nxg,nyg,nzg,ngrp_dif))

  ;Reading all the files to fill the variable structure
  ;****************************************************
  For i=1,ncpu do begin
     if ( i lt 10) then begin
        name='liste-0000'+strcompress(string(i),/remove_all)+out+'.Bin'
     endif else if (i lt 100) then begin
        name='liste-000'+strcompress(string(i),/remove_all)+out+'.Bin'
     endif else if (i lt 1000) then begin
        name='liste-00'+strcompress(string(i),/remove_all)+out+'.Bin'
     endif else if (i lt 10000) then begin
        name='liste-0'+strcompress(string(i),/remove_all)+out+'.Bin'
     endif else begin
        name='liste-'+strcompress(string(i),/remove_all)+out+'.Bin'
     endelse

     name = dir+'/'+name
     v = 0
     if (i eq 1) then v = 1
     
     text = 'Reading file: '+name
     slen=strlen(text)
     print,string(replicate(8b,slen(0))),f='($,A)'
     print,text,f='($,A)'
     
     rdat_str,name,var,verbose=v
  Endfor
  
  print,' '
  return
  
end

;*************************************************************************************
;
;*************************************************************************************

pro rdat_str,file,var,verbose=verbose

  openr,1,file,/f77_unformatted  
  ndim=1L 
  nx  =1L & ny =1L & nz  =1L
  nxg =1L & nyg=1L & nzg =1L
  ncpu=1L & geometry='x'
  nfx =1L
  nbuf=1L
  ngrp_ray=0L
  ngrp_dif=0L

  mhd=1L & rhd=1L & dif=1L & cond=1L & grav=1L

  xs  =1l & ys = 1L & zs = 1L
  xe  =1l & ye = 1L & ze = 1L
 
  readu,1,nxg,nyg,nzg,ncpu,geometry
  xg=dblarr(1,nxg+1) &   yg=dblarr(1,nyg+1) &   zg=dblarr(1,nzg+1)  
  readu,1, xg,yg,zg
  readu,1,ndim,nx,ny,nz,nfx,nbuf,ngrp_ray,ngrp_dif
  readu,1,mhd,rhd,dif,cond,grav
  readu,1,xs,ys,zs,xe,ye,ze

  if (keyword_set(verbose) ) then begin
    print,' '
    print,'ndim=',ndim
    print,'nx  =',nx
    print,'ny  =',ny
    print,'nz  =',nz
 
    print,'nfx =',nfx

    print,'ngrp_ray=',ngrp_ray
    print,'ngrp_dif=',ngrp_dif

    print,'mhd,rhd,dif,cond,grav=',mhd,rhd,dif,cond,grav
    
    print,'xs,xe = ',xs,xe
    print,'ys,ye = ',ys,ye
    print,'zs,ze = ',zs,ze
  endif	

  x = dblarr(nx+1)
  y = dblarr(ny+1)
  z = dblarr(nz+1)

  readu,1,x,y,z

   shift_gr= dblarr(ndim)
  nshift_gr= lonarr(ndim)
  readu,1, shift_gr
  readu,1,nshift_gr

  k = 1L
  readu,1,k

  time = 1D0 & dt = 1D0 & dtold = 1D0 
  readu,1,time, dt, dtold

  gamma = 1D0 & mu = 1D0
  readu,1,gamma,mu
  
  u_M = 1D0 & u_L = 1D0 & u_T = 1D0
  readu,1,u_M,u_L,u_T

  var.time=time & var.dt=dt & var.dtold=dtold & var.gamma=gamma & var.mu=mu & var.u_M=u_M & var.u_L=u_L & var.u_T=u_T & var.nite=k

  if (keyword_set(verbose) ) then begin
    print, 'Iteration =',k
    print, 'Time, dt,dtold = ',time, dt, dtold
    print, 'gamma    = ',gamma
    print, 'mu       = ',mu
    print, 'nfx      = ',nfx
    print, 'ngrp_ray = ',ngrp_ray
    print, 'ngrp_dif = ',ngrp_dif
  endif

  nbufx = nbuf
  nbufy = 0
  nbufz = 0
  if (ndim ge 2) then begin
     nbufy = nbuf
  endif
  if (ndim ge 3) then begin
     nbufz = nbuf
  endif

  rho  = dblarr(nx+2*nbufx,ny+2*nbufy,nz+2*nbufz)
  E    = dblarr(nx+2*nbufx,ny+2*nbufy,nz+2*nbufz)
  rhou = dblarr(nx+2*nbufx,ny+2*nbufy,nz+2*nbufz,ndim)
  if(nfx ge 1) then fx   = dblarr(nx+2*nbufx,ny+2*nbufy,nz+2*nbufz,nfx)
  readu,1,rho
  readu,1,rhou
  readu,1,E
  if(nfx ge 1) then readu,1,fx

  if rhd then begin	
      var.ngrp_ray=ngrp_ray
      er   = dblarr(nx+2*nbufx,ny+2*nbufy,nz+2*nbufz     ,ngrp_ray)
      fr   = dblarr(nx+2*nbufx,ny+2*nbufy,nz+2*nbufz,ndim,ngrp_ray)
      readu,1,er	
      readu,1,fr	
  endif	  

  if dif then begin	
      var.ngrp_dif=ngrp_dif
      ed   = dblarr(nx+2*nbufx,ny+2*nbufy,nz+2*nbufz,ngrp_dif)
      readu,1,ed	
  endif	  

  if mhd then begin	
      B   = dblarr(nx+2*nbufx,ny+2*nbufy,nz+2*nbufz,3)
      readu,1,B
  endif	  

  close,1

  kb  =   1.38062e-16 
  uma =  1.660531e-24 

  rhou2 = dblarr(nx+2*nbufx,ny+2*nbufy,nz+2*nbufz)
  for idim=0,ndim-1 do begin
     rhou2(*,*,*) = rhou2(*,*,*) + rhou(*,*,*,idim)^2
  endfor

  if mhd then begin	
      B2 = dblarr(nx+2*nbufx,ny+2*nbufy,nz+2*nbufz)
      for idim=0,ndim-1 do begin
         B2 = B2 + B(*,*,*,idim)^2
      endfor
      xx = E - 0.5*rhou2/rho - 0.5*B2
  endif else begin
      xx = E - 0.5*rhou2/rho
  endelse	  

  var.rho(xs-1:xe-1,ys-1:ye-1,zs-1:ze-1)=rho(nbufx:nx+nbufx-1,nbufy:ny+nbufy-1,nbufz:nz+nbufz-1)
  var.E  (xs-1:xe-1,ys-1:ye-1,zs-1:ze-1)=  E(nbufx:nx+nbufx-1,nbufy:ny+nbufy-1,nbufz:nz+nbufz-1)
  For idim=0,ndim-1 do begin
      var.u(xs-1:xe-1,ys-1:ye-1,zs-1:ze-1,idim)=rhou(nbufx:nx+nbufx-1,nbufy:ny+nbufy-1,nbufz:nz+nbufz-1,idim)/rho(nbufx:nx+nbufx-1,nbufy:ny+nbufy-1,nbufz:nz+nbufz-1)
  endfor
  
  ;Eint
  var.eint(xs-1:xe-1,ys-1:ye-1,zs-1:ze-1)= xx(nbufx:nx+nbufx-1,nbufy:ny+nbufy-1,nbufz:nz+nbufz-1)

  ;Pressure
  xx=(gamma-1)*xx
  var.p(xs-1:xe-1,ys-1:ye-1,zs-1:ze-1         )=xx(nbufx:nx+nbufx-1,nbufy:ny+nbufy-1,nbufz:nz+nbufz-1)

  ;Temperature
  xx = (xx/rho)*(mu*uma/kb)
  var.T  (xs-1:xe-1,ys-1:ye-1,zs-1:ze-1)= xx(nbufx:nx+nbufx-1,nbufy:ny+nbufy-1,nbufz:nz+nbufz-1)

  if(nfx ge 1) then var.fx(xs-1:xe-1,ys-1:ye-1,zs-1:ze-1,0:nfx-1 )=fx(nbufx:nx+nbufx-1,nbufy:ny+nbufy-1,nbufz:nz+nbufz-1,0:nfx-1)

  if rhd then begin	
      var.er(xs-1:xe-1,ys-1:ye-1,zs-1:ze-1         ,0:ngrp_ray-1)=er(nbufx:nx+nbufx-1,nbufy:ny+nbufy-1,nbufz:nz+nbufz-1         ,0:ngrp_ray-1)
      var.fr(xs-1:xe-1,ys-1:ye-1,zs-1:ze-1,0:ndim-1,0:ngrp_ray-1)=fr(nbufx:nx+nbufx-1,nbufy:ny+nbufy-1,nbufz:nz+nbufz-1,0:ndim-1,0:ngrp_ray-1)
  endif	  

  if dif then begin	
      var.ed(xs-1:xe-1,ys-1:ye-1,zs-1:ze-1         ,0:ngrp_dif-1)=ed(nbufx:nx+nbufx-1,nbufy:ny+nbufy-1,nbufz:nz+nbufz-1,0:ngrp_dif-1)
  endif	  

  if mhd then begin
      var.B(xs-1:xe-1,ys-1:ye-1,zs-1:ze-1,0:2         )=B(nbufx:nx+nbufx-1,nbufy:ny+nbufy-1,nbufz:nz+nbufz-1,0:2)
  endif	  

  return
end
